/*dib.c
* Copyright (C) 2008-2009 Antonio Maschio @ <tbin at libero dot it>
* 
* This is dib, the Dartmouth 1966 BASIC intepreter, a very naive
* attempt to implement the original Dartmouth College BASIC language as 
* conceived by Kemeny & Kurz in 1964, in its 1966 version (why 1966? 
* I was born in 1966). The interpreter is implemented in a not interactive 
* environment; it is in plain C (for the gcc), and it works as a common 
* UNIX command. See the file "dib_manual.txt" for explanation about the 
* interpreter and the language, and about its history and influences.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* It's GPL! Enjoy!
*
* (Sources should be read bottom-up, so if you read it top-down you will
* go from last complex functions to first simple and main)
* (Written with vim, tab stop = 8) */

#include "dib.h"


/* The following routines define the behavior of some peculiar 
 * functions added to dib as extended features and which were not 
 * part of the original Dartmouth BASIC. */


/* FIX()
 * return a number with decimals trimmed to the specified number; the digits 
 * are constrained into a well fixed range depending on the number;
 * FIX(number, 0) is equivalent to INT(number), and so is FIX(number, -10) 
 * This algorithm keeps in count the meaningful digits, so that
 * FIX(0.0003,2) returns 0 since in the first two decimal places 
 * of 0.0003 there are no meaningful digits */
double FIX(void) {
	int dec=0, digits=0, sign=1;

	/* get number and turn it to positive, storing sign */
	double num=S();
	if (num<0) num*=-1, sign=-1;

	/* get digits indicator */
	if (*p!=',') {perror_(5,l,false_); return 0.0;}
	p++;
	digits=S();

	/* FIX(num,0) is equivalent to INT(num) 
	 * the same for a negative number of decimals */
	if (digits <= 0) return (int)num;
	/* calculate ten-powers and subtract from digits */
	while (num < 1)  num*=10, dec++;
	while (num > 10) num/=10, dec--;
	digits-=dec;

	/* now 'num' is normalized and 'digits' holds the trimming value
	 * while 'dec' (positive or negative) still holds the ten-powers;
	 * in case 'dec' is negative, 1/pow(10,dec) = pow(10,-dec) */
	num=(round(num*pow(10,digits)))/pow(10,digits);
	return sign*num/pow(10,dec);
}


/* GAMMA()
 * The GAMMA function
 * return gamma value; argument is taken from stream 
 * Note: it would be cleaner to return lgamma() directly into getFunc;
 * I create this function because in the future I plan to implement
 * an algorithm on my own */
double GAMMA(void) {
	return signgam*exp(lgamma(S()));
}


/* CHANGE 
 * The CHANGE statement
 * convert the explicit string (enclosed into parentheses) to the ASCII
 * character values it's composed of, stored into the vector specified
 * after TO; the vector must not be already dimensioned, but if the
 * string is longer than 10 (with option base 1) or 11 (with option base 0)
 * an error will be raised.
 * Note: the original CHANGE statement worked even in the reverse way, that
 * is turning a vector of ASCII characters to a string variable, but
 * dib has no string variables and cannot enable this feature (yet). */
void CHANGE(void) {
	char *ps=p;	// points to string, terminated by double quote
	int var=0, i=ob, len=0;

	/* the string must be explicitly enclosed into double quotes;
	 * the following line must be changed if string vars are introduced */	
	if (*ps++!='\"') perror_(5,l,true_); 

	/* reach end of string*/
	p++; // skip double starting quotes
	while (*p!='\"') p++,len++;
	p++; // skip double closing quotes

	/* check syntax */
	if (strncmp(p,"TO",2)) perror_(5,l,true_);
	p+=2;
	
	/* get vector code */
	if (issv(p)) var=*p;
	else if (isdv(p)) var=in(*(p),*(p+1));
	else perror_(6,l,true_);

	/* check vector existence and dimension */
	if (dim[var].type==2) perror_(20,l,true_);
	else if (dim[var].type!=1) dimArray(var,1,10,0);
	
	/* do the string conversion, storing in position 0 (or 1
	 * if option base is 1) the string length */
	setArrayVal(var,1,i++,0,len);
	while (*ps!='\"') setArrayVal(var,1,i++,0,*ps++);
}


/* The following routines define the behavior of the MAT statements added
 * to dib as extended features and which were not part of the original
 * Dartmouth BASIC.
 * Ideas for some of them were taken from the HP-71B pocket computer BASIC
 * language and features */


/* DET
 * The DET function
 * by calling matinvert() on a temporary matrix, 
 * set global variable detValue as the determinant */
void MATDET(void) {
	int var1;

	/* get matrix code */
	if (issv(p)) var1=*p++;
	else if (isdv(p)) var1=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return ;}

	/* check dimensions validity
	 * only square matrices have a determinant */
	if (dim[var1].type != 2 || dim[var1].row != dim[var1].col)
		perror_(20,l,true_); 

	/* create dummy matrix dimensionally equal to var1 */
	dimArray(TEMPVAR,2,dim[var1].row,dim[var1].col);

	/* This call to matinvert calculates the inverted matrix and
	 * stores it into TEMPVAR, while setting detValue to the determinant 
	 * (or exit with error in case of singularity). 
	 * The inverted matrix of TEMPVAR is not used, because we are only 
	 * interested in detValue. */
	matinvert(TEMPVAR,var1);
}


/* MATNORM()
 * The FNORM, RNOMR and CNORM functions
 * return value calculated in normer() */
double MATNORM(void) {
	int var1, type=0;
	
	if (*p == 'F') {
		p+=6;
		type = FNORMTYPE;
	}
	else if (*p == 'R') {
		p+=6;
		type = RNORMTYPE;
	}
	else if (*p == 'C') {
		p+=6;
		type = CNORMTYPE;
	}
	
	/* get argument matrix code */
	if (issv(p)) var1=*p++;
	else if (isdv(p)) var1=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return 0;}

	return normer(var1, type);
}


/* normer(argument-matrix, norm-type)
 * return Frobenius norm, row norm or column norm of
 * argument-matrix depending on norm-type */
double normer(int var, int type) {
	int r=dim[var].row;
	int c=dim[var].col;
	int t=dim[var].type;
	int i,j;
	double norm=0;

	/* Frobenius norm */
	if (type == FNORMTYPE) {
		double v=0;
		i=ob; j=(t==1)?0:ob;
		while (i<=r) {
			v=getArrayVal(var,t,i,j);
			norm += v*v;
			j++;
			if (j>c) {i++; j=(t==1)?0:ob;}
		}
		norm = pow(norm, 0.5);
	}
	/* Row norm */
	else if (type == RNORMTYPE) {
		double max=0;
		i=ob; 
		while (i<=r) {
			j=(t==1)?0:ob;	
			norm=0;
			while (j<=c) {
				norm += fabs(getArrayVal(var,t,i,j));
				j++;
			}
			if (max < norm) max = norm;
			i++;
		}
		norm = max;
	}
	/* Column norm */
	else if (type == CNORMTYPE) {
		double max=0;
		j=(t==1)?0:ob;
		while (j<=c) {
			norm=0;
			i=ob;
			while (i<=r) {
				norm += fabs(getArrayVal(var,t,i,j));
				i++;
			}
			if (max < norm) max = norm;
			j++;
		}
		norm = max;
	}
	return norm;
}


/* MATDOT()
 * The DOT statement
 * return dot product between two unidimensional vectors, 
 * yielding a real argument vectors are taken from stream
 * Vectors may be vertical or horizontal in either side
 * This algorithm is a bit clumsy, I know */
double MATDOT(void) {
	int var1,var2;
	double res=0.0;
	int dim1=0, dim2=0;

	/* get matrices codes */
	if (issv(p)) var1=*p++;
	else if (isdv(p)) var1=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return 0;}
	if (*p++!=',') {perror_(2,l,false_); return 0;}
	if (issv(p)) var2=*p++;
	else if (isdv(p)) var2=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return 0;}

	/* check types */
	if (dim[var1].row>ob && dim[var1].col>ob) perror_(20,l,true_); 
	if (dim[var2].row>ob && dim[var2].col>ob) perror_(20,l,true_); 
	
	/* get dimensions */
	if (dim[var1].row!=0 && dim[var1].col<=ob)      dim1=dim[var1].row;
	else if (dim[var1].row<=ob && dim[var1].col!=0) dim1=dim[var1].col;
	if (dim[var2].row!=0 && dim[var2].col<=ob)      dim2=dim[var2].row;
	else if (dim[var2].row<=ob && dim[var2].col!=0) dim2=dim[var2].col;
	
	/* check dimensions validity */
	if (dim1!=dim2) perror_(20,l,true_);

	/* perform dot product */
	else {
		double vec1[dim1+1], vec2[dim1+1];
		int i;
		
		/* copy var1 from either type */
		if (dim[var1].type==1) {
			for (i=ob;i<=dim1;i++)
				vec1[i]=getArrayVal(var1,1,i,0);
		}	
		else {
			if (dim[var1].row!=ob && dim[var1].col==ob)
				for (i=ob;i<=dim1;i++) 
					vec1[i]=getArrayVal(var1,2,i,ob);
			else if (dim[var1].row==ob && dim[var1].col!=ob) 
				for (i=ob;i<=dim1;i++) 
					vec1[i]=getArrayVal(var1,2,ob,i);
		}
		/* copy var 2 from either type */
		if (dim[var2].type==1) {
			for (i=ob;i<=dim1;i++)
				vec2[i]=getArrayVal(var2,1,i,0);
		}
		else {
			if (dim[var2].row!=ob && dim[var2].col==ob)
				for (i=ob;i<=dim2;i++) 
					vec2[i]=getArrayVal(var2,2,i,ob);
			else if (dim[var2].row==ob && dim[var2].col!=ob) 
				for (i=ob;i<=dim2;i++) 
					vec2[i]=getArrayVal(var2,2,ob,i);
		}

		/* calculate dot product from copies */
		for (i=ob;i<=dim1;i++) res+=vec1[i]*vec2[i];
	}
	return res;
}


/* The following routines define the behavior of the MAT statements which 
 * were included as standard on the 1966 version of the Dartmouth BASIC,
 * and as a CardBasic option in version II of 1964.
 * Being the October 1964 manual for version II quite poor in presenting 
 * these tools, and practically useless for what about code examples, I must 
 * rely on something else, and precisely on the first version of the book
 * "Programming with BASIC" by Byron S. Gottfried (1975, Italian edition)
 * and on the 1978 historical essay "BASIC Session" by Thomas E. Kurtz.
 * MATINPUT, which was not in the 1966 III version but in the 1967 IV version,
 * is enabled in dib as a "bonus" */


/* MATbyK(result-matrix)
 * multiply a matrix by an expression, both taken from stream */
void MATbyK(int varr){
	double exp=S();
	int var1, i,j,m,n,t;

	/* check syntax */
	if (*p++!= ')') {perror_(7,l,false_); return;}
	if (*p++!= '*') {perror_(7,l,false_); return;}
	
	/* get argument matrix code */
	if (issv(p)) var1=*p++;
	else if (isdv(p) && ef) var1=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return;}

	/* check matrices existence */
	if (dim[varr].type==0 || dim[var1].type==0) perror_(20,l,true_);

	/* check dimensions validity */
	if (dim[varr].row!=dim[var1].row || dim[varr].col!=dim[var1].col)
		perror_(20,l,true_);

	/* enable MAT A=(k)*A only for extended features */
	if (!ef && var1==varr) perror_(20,l,true_);

	/* perform multiplication by expression */
	m=dim[varr].row;
	n=dim[varr].col;
	t=dim[varr].type;
	i=ob; j=(t==1)?0:ob;
	while (i<=m) {
		setArrayVal(varr,t,i,j,exp*getArrayVal(var1,t,i,j));
		j++;
		if (j>n) {i++; j=(t==1)?0:ob;}
	}
}


/* MATTRN(result-matrix)
 * The MAT TRN statement
 * Read argument matrix from stream, transpose it
 * and set result-matrix as the transposed matrix */
void MATTRN(int varr){
	int var1,i,j,t,t1;

	if (*p++=='(') {
		/* get argument matrix code */
		if (issv(p)) var1=*p++;
		else if (isdv(p) && ef) var1=in(*(p),*(p+1)),p+=2;
		else {perror_(6,l,false_); return;}
	}
	else {perror_(7,l,false_); return;}

	/* check equality */
	if (!ef && varr==var1) perror_(20,l,true_);

	/* check dimensions validity */
	if (dim[varr].row!=dim[var1].col || dim[varr].col!=dim[var1].row)
		perror_(20,l,true_);

	/* dimension temporary matrix as result matrix */
	t=dim[varr].type;
	t1=dim[var1].type;
	dimArray(TEMPVAR,t,dim[varr].row,dim[varr].col);
	
	/* transpose argument matrix over temporary; 
	 * I do this to enable MAT A=TRN(A) in extended mode */
	i=(t==1)?0:ob;
	while (i<=dim[varr].col) {
		j=ob;
		while (j<=dim[varr].row) {
			setArrayVal(TEMPVAR,t,j,i,getArrayVal(var1,t1,i,j));
			j++;
		}
		i++;
	}

	/* copy back to result matrix */
	for(i=0;i<=dim[varr].row;i++) for(j=0;j<=dim[varr].col;j++)
		setArrayVal(varr,t,i,j,getArrayVal(TEMPVAR,t,i,j));
}


/* MATINV(result-matrix)
 * The MAT INV statement
 * Read parameters from the stream, setup parameters 
 * and pass them to matinvert(), the real matrix inverter. */
void MATINV(int varr){
	int var1;

	if (*p++=='(') {
		/* get argument matrix code */
		if (issv(p)) var1=*p++;
		else if (isdv(p) && ef) var1=in(*(p),*(p+1)),p+=2;
		else {perror_(6,l,false_); return;}
	}
	else {perror_(6,l,false_); return;}

	/* check matrix existence */
	if (dim[varr].type!=2 || dim[var1].type!=2) perror_(20,l,true_);

	/* check equality 
	 * MAT A=INV(A) is available only as an extended feature */
	if (!ef && varr==var1) perror_(20,l,true_); 

	/* check dimensions validity (only square matrices may be inverted) */
	if (dim[varr].row!=dim[varr].col || dim[var1].row!=dim[var1].col\
			|| dim[varr].row!=dim[var1].row\
			|| dim[varr].col!=dim[var1].col)
		perror_(20,l,true_);

	/* delegate inversion */
	matinvert(varr,var1);

	/* set det flag after calculation */
	isdetCalc=TRUE;
}


/* matinvert(result-matrix, argument-matrix)
 * Inversion of matrix 'a' into matrix 'x' and calculation of the determinant
 * using the Gauss factorization scheme. Variables:
 * - varr is the unknown global inverted matrix (NxN) in the BASIC code
 * - var1 is the global knowns matrix to be inverted (NxN) in the BASIC code; 
 *   it's used to form ma
 * - ma is the local composed matrix (2NxN) (to the left the knowns matrix, 
 *   to the right the identity matrix); this matrix is invisible to BASIC
 * - x is the local calculated inverted matrix (NxN), invisible to BASIC
 *   and later copied back to varr. 
 * - detValue is the global determinant value, visible to BASIC through DET
 * Copyright Antonio Maschio - March 2008 */
void matinvert(int varr, int var1) {
	int n = (dim[var1].row), d=2*n+1; 	
	int i, j, k;
	double x[n+1][n+1],ma[n+1][d+1];
	double mul,piv;

	/* set determinant multiplier */
	detValue=1.0;
	
	/* create composed matrix AI, with the same rows number of A,
	 * and twice the columns number of A; in the second half lies I */
	for (i=ob;i<=n;i++) for(j=ob;j<=n;j++) {
		ma[i][j]=getArrayVal(var1,2,i,j);
		/* setting I to all zeroes */
		ma[i][j+n+1]=0.0;
	}
	/* set diagonal of I to all ones */
	for (i=ob;i<=n;i++) ma[i][i+n+1]=1.0;
	
	/* erase destination matrix */
	for (i=ob; i<=n; i++) for (j=ob; j<=n; j++) x[i][j]=0.0;
	
	/* Gauss factorization over AI with partial pivoting */
	for (k=ob;k<n;k++) {
		piv=ma[k][k];
		if (!piv) {
			double maxx=0;
			int r=k;
			/* k is the row where pivot is zero
			 * now find the row where pivot is max */
			for (i=k+1;i<=n;i++) 
				if (maxx<fabs(ma[i][k])){ 
					maxx=ma[i][k]; 
					r=i; 
				}
			/* matrix is singular if r hasn't changed */
			if (r==k) { perrorEx_(41,l); detValue=0; return; }
			/* exchange rows r and k */
			for (i=ob;i<=d;i++) {
				maxx=ma[r][i];
				ma[r][i]=ma[k][i];
				ma[k][i]=maxx;
			}
			piv=ma[k][k];
			/* matrix is singular if pivot is still null */
			if (!piv) { perrorEx_(41,l); detValue=0; return; }
			/* update determinant sign */
			detValue*=-1; 
		}
		/* update the rest of the rows */
		for (i=k+1;i<=n;i++) {
			mul=ma[i][k]/piv;
			for (j=k;j<=d;j++) ma[i][j]=ma[i][j]-mul*ma[k][j];
		} 
		
	}

	/* calculate determinant */
	for (i=ob;i<=n;i++) detValue*=ma[i][i];
	
	/* check singularity (this should be useless, 
	 * but it's safer checking it the same) */
	if (detValue==0.0) { perrorEx_(41,l); return; }

	/* calculate the inverse of A through back-substitution */
	for (i=ob;i<=n;i++) {
		x[n][i]=ma[n][i+n+1]/ma[n][n];
		for (j=n-1;j>=0;j--) {
			piv=0;
			for(k=j+1;k<=n;k++) piv+=ma[j][k]*x[k][i];
			x[j][i]=(ma[j][i+n+1]-piv)/ma[j][j];
		}
	}

	/* set varr to the resulting inverted matrix */
	for (i=0; i<=n; i++) for (j=0; j<=n; j++) 
		setArrayVal(varr,2,i,j,x[i][j]);
}


/* MATIDN(result-matrix)
 * The MAT IDN statement
 * set a square array to identity matrix */
void MATIDN(int var){
	int m=0, i, j;

	/* check existence */
	if (dim[var].type==0) perror_(20,l,true_);

	/* IDN is only for square matrices */
	if (dim[var].col!=dim[var].row) perror_(20,l,true_);

	/* get redimensioning data */
	if (*p=='(') {
		p++;
		m=(int)S();
		if (m<0) perror_(20,l,true_);
		if (0==m && ob) perror_(20,l,true_);
		if (*p++!=')') {perror_(7,l,false_); return;}
		
		/* redimension */
		if (m<=dim[var].row) dim[var].row=dim[var].col=m;
		else perror_(20,l,true_);
	}
	/* set existing dimensioning data */
	else if (*p=='\0') m=dim[var].row;
	else {perror_(7,l,false_); return;}

	/* set values for IDN */
	for (i=ob;i<=m;i++) for (j=ob;j<=m;j++) setArrayVal(var,2,i,j,0.0);
	for (i=ob;i<=m;i++) setArrayVal(var,2,i,i,1.0);
}


/* MATZER(result-matrix, option-value)
 * The MAT ZER and MAT CON statements
 * set (or reset) an array to zero (MAT ZER) or one (MAT CON)
 * depending on option value 'op' (0 for ZER, 1 for CON); */
void MATZER(int var, double op){
	int t=0, m=0, n=0;
	int i,j;

	/* check existence */
	if (dim[var].type==0) perror_(20,l,true_);

	/* get redimensioning data */
	if (*p=='(') {
		p++;
		m=(int)S();
		if (m<0) perror_(20,l,true_);
		if (0==m && ob) perror_(20,l,true_);
		/* second index? bidimensional array */
		if (*p==',') {
			if (dim[var].type==1) perror_(29,l,true_);
			p++;
			n=(int)S();
			if (n<0) perror_(20,l,true_);
			if (0==n && ob) perror_(20,l,true_);
		}
		else if (dim[var].type==2) perror_(20,l,true_);

		/* redimension */ 
		if (m<=dim[var].row) dim[var].row=m;
		else perror_(20,l,true_);
		if (dim[var].type==2) {
			if (n<=dim[var].col) dim[var].col=n;
			else perror_(20,l,true_);
		}
		if (*p++!=')') {perror_(7,l,false_); return;}
	}
	/* set existing dimensioning data */
	else if (*p=='\0') {
		m=dim[var].row;
		n=dim[var].col;
	}
	else {perror_(7,l,false_); return;}


	/* set values for ZER or CON */
	t=dim[var].type;
	i=ob; j=(t==1)?0:ob;
	while (i<=m) {
		setArrayVal(var,t,i,j,op);
		j++;
		if (j>n) {i++; j=(t==1)?0:ob;}
	}
}


/* MATsum(result-matrix,matrix1,matrix2,option-value)
 * Add two matrices or subtract the second from the first;
 * option value is -1 for subtractions, +1 for additions */
void MATsum(int varr, int var1, int var2, int op){
	int i,j;
	int M=dim[varr].row;
	int N=dim[varr].col;
	int T=dim[varr].type;
	int T1=dim[var1].type;
	int T2=dim[var2].type;

	/* create dummy matrix dimensionally equal to varr */
	dimArray(TEMPVAR,T,M,N);
	
	/* check tables existence */
	if (T==0 || T1==0 || T2==0) perror_(20,l,true_);

	/* check matching of dimensions 
	 * Note: one can only sum vectors with vectors, matrices with
	 * matrices, and not use unidimensional vectors or
	 * degenerated matrices of one row or of one column */
	if ((M!=dim[var1].row) || M!=dim[var2].row\
			|| N!=dim[var1].col || N!=dim[var2].col)
		perror_(20,l,true_);

	/* enable MAT A=A+B or MAT A=B+A or MAT A=A+A 
	 * only for extended features */
	if (!ef && (varr==var1 || varr==var2)) perror_(20,l,true_);
	
	/* perform addition/subtraction on temp var 
	 * and copy back to varr - this is done to enable 
	 * MAT A=A+B or MAT A=B+A or MAT A=A+A */
	i=ob; j=(T==1)?0:ob;
	while (i<=M) {
		setArrayVal(TEMPVAR,T,i,j,\
				getArrayVal(var1,T1,i,j)+\
				op*getArrayVal(var2,T2,i,j));
		setArrayVal(varr,T,i,j,getArrayVal(TEMPVAR,T,i,j));
		j++;
		if (j>N) {i++; j=(T==1)?0:ob;}
	}

}


/* MATmul(result-matrix,matrix1,matrix2)
 * The cross-product between two matrices or vectors */
void MATmul(int varr, int var1, int var2){
	int i,j,k;
	int M=dim[varr].row;
	int Nr=dim[varr].col;
	int N=dim[var1].col;
	int P=dim[var2].col;
	int T=dim[varr].type;
	int T1=dim[var1].type;
	int T2=dim[var2].type;

	/* instantiate temporary matrix with result matrix dimensions */
	dimArray(TEMPVAR,T,M,Nr);

	/* check existence */
	if (T==0 || T1==0 || T2==0) perror_(20,l,true_);

	/* check dimensions validity */
	if ((dim[var1].col!=dim[var2].row) || dim[varr].row!=dim[var1].row\
			|| dim[var2].col!=dim[varr].col)
		perror_(20,l,true_);

	/* enable MAT A=A*B or MAT A=B*A or MAT A=A*A
	 * only for extended features */
	if (!ef && (varr==var1 || varr==var2)) perror_(20,l,true_);

	/* perform cross multiplication and copy data back to varr;
	 * this is done to enable MAT A=A*B or MAT A=B*A or MAT A=A*A */
	if (T==2) for(i=ob;i<=M;i++) {
		for(j=ob;j<=P;j++) {
			double res=0.0;
			for(k=ob;k<=N;k++)
				res+=getArrayVal(var1,T1,i,k)*\
					getArrayVal(var2,T2,k,j);
			setArrayVal(TEMPVAR,T,i,j,res);
			setArrayVal(varr,T,i,j,getArrayVal(TEMPVAR,T,i,j));
		}
	}
	else if (T==1) for(i=ob;i<=M;i++) {
		double res=0.0;
		for(k=ob;k<=N;k++)
			res+=getArrayVal(var1,T1,i,k)*\
				getArrayVal(var2,T2,k,0);
		setArrayVal(TEMPVAR,T,i,0,res);
		setArrayVal(varr,T,i,0,getArrayVal(TEMPVAR,T,i,0));
	}
}


/* MATtype()
 * this function is invoked by RUN when a MAT statement is met
 * (other than MAT READ, MAT PRINT or MAT INPUT, which follow their way) 
 * to choose the correct command. */
void MATtype(void) {
	int varr=0; 
	int var1=0;
	int var2=0;
	int op=0;
	
	/* get result matrix code */
	if (issv(p)) varr=*p++;
	else if (isdv(p) && ef) varr=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return;}
	
	if (*p++!='=') {perror_(7,l,false_); return;} 

	/* mat operations */
	if (issv(p) || isdv(p)) {
		/* get first matrix operand code */
		if (issv(p)) var1=*p++;
		else if (isdv(p) && ef) var1=in(*(p),*(p+1)),p+=2;
		else {perror_(6,l,false_); return;}
		
		/* get op */
		if (*p!='+' && *p!='-' && *p!='*' && *p!='\0') 
			{perror_(7,l,false_); return;}
		else op=*p++;
		
		/* get second matrix operand code */
		if (op!='\0') {
			if (issv(p)) var2=*p++;
			else if (isdv(p) && ef) var2=in(*(p),*(p+1)),p+=2;
			else {perror_(6,l,false_); return;}
		}
		
		/* perform operations */
		     if (op=='-') MATsum(varr, var1, var2, -1);
		else if (op=='+') MATsum(varr, var1, var2, 1);
		else if (op=='*') MATmul(varr, var1, var2);
		else if (op=='\0') { 
			if (ef) matCopy(varr, var1);
			else {perror_(7,l,false_); return;}
		}
		else {perror_(7,l,false_); return;}
	}
		
	/* mat functions */
	else {
		     if (!strncmp(p,"ZER",3)) p+=3, MATZER(varr,0.0);
		else if (!strncmp(p,"CON",3)) p+=3, MATZER(varr,1.0);
		else if (!strncmp(p,"TRN",3)) p+=3, MATTRN(varr);
		else if (!strncmp(p,"IDN",3)) p+=3, MATIDN(varr);
		else if (!strncmp(p,"INV",3)) p+=3, MATINV(varr);
		else if (*p=='(') p++, MATbyK(varr);
		else perror_(7,l,false_);
	}
}


/* matCopy(result-matrix, argument-matrix)
 * copy argument matrix into result matrix
 * extended mode only - the check is done into MATtype() */
void matCopy(int dst, int src){
	int ms=dim[src].row;
	int ns=dim[src].col;
	int ts=dim[src].type;
	int MD=dim[dst].row;
	int ND=dim[dst].col;
	int TD=dim[dst].type;
	int i,j;

	/* check compatibility */
	if (dst==src) return;
	if (ms!=MD || ns!=ND || ts!=TD || ts==0 || TD==0) perror_(20,l,true_);

	if (ts==2) for (i=0;i<=ms;i++) for (j=0;j<=ns;j++)
		setArrayVal(dst,TD,i,j,getArrayVal(src,ts,i,j));
	else if (ts==1) for (i=0;i<=ms;i++)
		setArrayVal(dst,TD,i,0,getArrayVal(src,ts,i,0));
}


/* matReadAssign(result-matrix,type,rows,columns)
 * iterate through matrix indexes to assign DATA values */
void matReadAssign(int var, int t, int m, int n){
	int i,j;
	/* vectors */
	if (1==t) {
		for(i=ob;i<=m;i++) {
			if (dc>dataLimit){
				/* if out of data, exit stopping the
				 * program (cfr. Gottfried cit. 7.2) */
				if (ef) perrorEx_(22,l); break;
				return;
			}
			setArrayVal(var,1,i,0,D[dc++]);
		}
	}
	/* matrices */
	else {
		for(i=ob;i<=m;i++) 
			for (j=ob;j<=n;j++) {
				if (dc>dataLimit){
					/* if out of data, exit stopping the
					 * program (cfr. Gottfried cit. 7.2) */
					if (ef) perrorEx_(22,l); break;
					return;
				}
				setArrayVal(var,2,i,j,D[dc++]);
			}
	}
}


/* MATINPUT()
 * The MAT INPUT statement
 * Note: this statement was introduced in the IV version of Dartmouth BASIC,
 * and it features here as a bonus.
 * Multiple-matrices input enabled */
void MATINPUT(void){
	double tp[LIMIT];
	char line[STRLIM+1],linens[STRLIM+1];
	char *dp= line, *pp=p;
	int cnt=0, clim=0, var=0, m=0,n=0,t=0, i=ob, j=ob;
	
	p=linens;
	printf("? ");
	fgets(line,STRLIM,stdin);
	line[strlen(line)-1]='\0';
	while (*dp) {
		if(*dp=='&') *dp='?'; // change to '?' to deceive S()
		if(*dp!=' ' && *dp!='\t') *p++=*dp; 
		dp++;	
	}
	*p='\0'; p=linens;

	/* store all values */
	while (*p){
		tp[cnt++]=S();
		if (*p=='\0') break;
		else if (*p=='?') {
			printf("? ");
			fgets(line,STRLIM,stdin);
			line[strlen(line)-1]='\0';dp=line;p=linens;
			while (*dp) {if(*dp!=' '&&*dp!='\t') *p++=*dp; dp++;}
			*p='\0'; p=linens;
		}
		else if(*p==',') p++;
		else {perror_(7,l,false_); return;}
	}
	clim=cnt-1;

	/* attribution of all values to arrays */
	cnt=0; 
	while (*pp) {
		numValue=0;
		i=ob;j=ob;

		/* get code for variable under assignment */
		if (issv(pp)) var=*pp++;
		else if (isdv(pp) && ef) var=in(*(pp),*(pp+1)),pp+=2;
		else {perror_(6,l,false_); return;}
		
		/* check existence, or else get matrix parameters */
		if ((t=dim[var].type)==0) perror_(20,l,true_);
		m=dim[var].row;
		n=dim[var].col;

		/* set values for variable in turn */
		while (i<=m) {
			if (cnt>clim) break;
			setArrayVal(var,t,i,j,tp[cnt++]);
			j++; numValue++;
			if (j>n) j=ob, i++; 
		}
		if (*pp++!=',') break;
	}
}


/* MATREAD()
 * The MAT READ statement 
 * Multiple-matrices reading enabled */
void MATREAD(void){
	int var=0, m=0, n=0;

	if (dataLimit==0) {
		perror_(10,l,false_);
		return;
	}
	/* start loop on all MAT READ arrays */	
	do {
		/* get variable code */
		if (issv(p)) var=*p++;
		else if (isdv(p) && ef) var=in(*(p),*(p+1)),p+=2;
		else {perror_(6,l,false_); return;}

		/* check existence */
		if (dim[var].type==0) perror_(20,l,true_);
	
		/* get explicit data */
		if (*p=='(') {
			p++;
			m=(int)S();
			if (m<0) perror_(20,l,true_);
			if (0==m && ob) perror_(20,l,true_);

			/* second index? bidimensional array */
			if (*p==',') {
				if (dim[var].type==1) perror_(20,l,true_);
				p++;
				n=(int)S();
				if (n<0) perror_(20,l,true_);
				if (0==n && ob) perror_(20,l,true_);
			}
			else if (dim[var].type==2) perror_(20,l,true_);

			/* redimension */ 
			if (m<=dim[var].row) dim[var].row=m;
			else perror_(20,l,true_);
			if (dim[var].type==2) {
				if (n<=dim[var].col) dim[var].col=n;
				else perror_(20,l,true_);
			}
			if (*p++!=')') {perror_(7,l,false_); return;}
		}

		/* set implicit data */
		else if (*p=='\0' || *p==',') {
			m=dim[var].row;
			n=dim[var].col;
		}
		else {perror_(7,l,false_); return;}

		/* assign DATA values */
		matReadAssign(var,dim[var].type,m,n);


	} while (*p++==',');
}


/* MATPRINT()
 * The MAT PRINT statement 
 * Multiple-matrices printing enabled */
void MATPRINT(void){
	int var=0, m=0, n=0, t=0;
	/* get variable code */
	if (issv(p)) var=*p++;
	else if (isdv(p) && ef) var=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return;}

 	m=dim[var].row;
	n=dim[var].col;
	t=dim[var].type;

	/* check if array was dimensioned */
	if (t!=1 && t!=2) perror_(20,l,true_);
	
	if (1==t) {
		int i;
		for (i=ob;i<=m;i++){
			counter=printNUM(getArrayVal(var,1,i,0));
			if (counter > SCRLIM) CR();
			if (*p==';') {
				/* advance to next minizone, with 
				 * 6 char minimum, and if too close, 
				 * go to next */
				int d=6-(counter-cw);
				while (d<2) d+=3;
				counter+=d;
				while (d-- > 0) putchar(' ');
				if (counter>SCRLIM-10) CR();
			}
			else if (*p==',') {
				while (counter%ZONE) AC();
				if (counter>SCRLIM) CR();
			}
			else if (*p=='\0') CR();
			else perror_(7,l,false_);
		}
	}
	else {
		int i,j;
		for (i=ob;i<=m;i++) {
			for(j=ob;j<=n;j++) {
				int cw=counter;
				counter=printNUM(getArrayVal(var,2,i,j));
				if (counter > SCRLIM) CR();
				if (*p==';') {
					/* advance to next minizone, with 
					 * 6 char minimum, and if too close, 
					 * go to next */
					int d=6-(counter-cw);
					while (d<2) d+=3;
					counter+=d;
					while (d-- > 0) putchar(' ');
					if (counter>SCRLIM-10) CR();
				}
				else if (*p==',' || *p=='\0')
					while (counter%ZONE) AC();
				else perror_(7,l,false_);
				if (counter>SCRLIM) CR();
			}
			CR();
			CR();
		}
	}
	CR();
	/* if there's something else to print after advancing p
	 * (and if this new p is not null) iterate recalling itself */
	if (*p) p++;
	if (*p) MATPRINT();
}


/* The following routines define the behavior of the DIM statement, which 
 * was included as standard since the 1964 version of the Dartmouth BASIC. */


/* isod(string-position)
 * (IS a One Dimensional array)
 * test if the current string points to a one (TRUE) or two (FALSE) 
 * dimensions array
 * p (which is not touched) points to the character after the opening '('; 
 * To test eventually for a bidimensional array, use !isod() */
int isod(char *p){
	char *h=p-1;
	while(')'!=*h && '\0'!=*h) if(','==*h++) return FALSE;
	return TRUE;
}


/* isaa(string-position)
 * (IS an Array Assignment)
 * test if the current string **begins** with an array assignment
 * (a feature enabled only for extended features) */
int isaa(char *p) {
	char *h=p;
	if (issv(h)) h++;
	else if (isdv(h) && ef) h+=2;
	else return FALSE;

	if ('('!=*h) return FALSE;
	while (')'!=*h) h++;
	if ('\0'==*h) return FALSE;
	h++;
	if ('='!=*h) return FALSE;
	return TRUE;
}	


/* dimArray(matrix-index,type,rows-number,columns-number)
 * dimension an array allocating the necessary memory space; Variables:
 * - var is the index of the variable holding the matrix name
 * - M is the number of rows (=0 for one rows matrices)
 * - N is the number of columns (=0 for vectors and one columns matrices)
 * - T is the type, 1 for unidimensional vectors, 2 for matrices
 *   The one-dimensional vector is supposed to be vertical (one column);
 *   controls over M&N<=0 are supposed to happen elsewhere 
 *   All elements of declared array are set to 0.0 */
void dimArray(int var, int T, int M, int N) {
	int i;
	int size1=sizeof(double)*((M+1)*2+1);
	int size2=sizeof(double)*((M+1)*(N+1)+1);
	dim[var].row  = M;
	dim[var].col  = N;
	dim[var].type = T;
	
	/* reserve memory space for the array */
	if (T==1) dim[var].elem=malloc(size1);  
	else if (T==2) dim[var].elem=malloc(size2);
	else {perror_(7,l,false_); return;}
	if (!dim[var].elem) 
		derror_("dib: out of memory for arrays.\n");

	/* erase all array elements */
	if (1==T) for(i=0;i<size1;i++) *(dim[var].elem+i)=0;
	else for(i=0;i<size2;i++) *(dim[var].elem+i)=0;
}		


/* getArrayVal(matrix-index,type,row-position,column-position)
 * get value of an element of an array, given its coordinates */
double getArrayVal(int var, int t, int m, int n) {
	if (1==t) {
		if (m<=10 && 0==dim[var].row && 0==dim[var].col) 
			dimArray(var,1,10,0);
		else if (m>10 && 0==dim[var].row) perror_(20,l,true_);
		else if (m>10 && dim[var].row<m) perror_(20,l,true_);
		else if (dim[var].row<m) perror_(20,l,true_);
		return *(dim[var].elem+m*(dim[var].col+1)+2);
	}
	else if (2==t) {
		if (m<=10 && 0==dim[var].row && n<=10 && dim[var].col==0) 
			dimArray(var,2,10,10);
		else if (0==dim[var].row && ob) perror_(20,l,true_);
		else if (dim[var].row<m || dim[var].col<n) perror_(20,l,true_);
		return *(dim[var].elem+(dim[var].col+1)*m+n); 
	}
	else perror_(20,l,true_);
	return 0;
}


/* setArrayVal(matrix-index,type,row-position,column-position,value)
 * set an element of an array to a specific value, 
 * given the value itself and the element coordinates */
void setArrayVal(int var, int t, int m, int n, double val){
	if (1==t) {
		if (m<=10 && 0==dim[var].row && 0==dim[var].col) 
			dimArray(var,1,10,0);
		else if (m>10 && 0==dim[var].row) perror_(20,l,true_);
		else if (m>10 && dim[var].row<m) perror_(20,l,true_);
		else if (dim[var].row<m) perror_(20,l,true_);
		*(dim[var].elem+m*(dim[var].col+1)+2)=val;
	}
	else if (2==t) {
		if (m<=10 && 0==dim[var].row && n<=10 && dim[var].col==0) 
			dimArray(var,2,10,10);
		else if (0==dim[var].row && ob) perror_(20,l,true_);
		else if (dim[var].row<m || dim[var].col<n) perror_(20,l,true_);
		*(dim[var].elem+(dim[var].col+1)*m+n)=val;
	}
	else perror_(20,l,true_);
	return;
}


/* dimAssign()
 * wrapper for setArrayVal to be called into LET()
 * for figures like A(N)=Expr1 or A(N,M)=Expr2 
 * As stated at page 36 of the 1964 manual for version II:
 * "The name of a list or a table must be a single letter." 
 * I allow the usage of A0..Z9 as an extended feature */
void dimAssign(void) {
	int var=0, m=0, n=0;
	if (issv(p)) var=*p++;
	else if (isdv(p) && ef) var=in(*(p),*(p+1)),p+=2;
	else {perror_(6,l,false_); return;}

	if ('('!=*p++) {perror_(7,l,false_); return;}
	m=(int)S();
	if (m<0) perror_(20,l,true_);
	if (m==0 && ob) perror_(20,l,true_);

	/* unidimensional array */
	if (*p==')') {
		p++; // skip ')'
		if (*p++!= '=') {perror_(7,l,false_); return;}
		setArrayVal(var,1,m,ob,S());
	}
	/* bidimensional array */
	else if (*p==',') {
		p++; // skip ','
		n=(int)S(); 
		if (n<0) perror_(20,l,true_);
		if (n==0 && ob) perror_(20,l,true_);
		if (')'!=*p) {perror_(7,l,false_); return;}
		p++; // skip ')'
		if ('='!=*p) {perror_(7,l,false_); return;}
		p++; // skip '='
		setArrayVal(var,2,m,n,S());
	}
	else {perror_(7,l,false_); return;}
}


/* setINPUTarray(string_pos, value) 
 * on an INPUT line, if there's an array var, get its references and 
 * stores the user given value; pos is the position in the program line, 
 * value is the value already retrieved by user input line 
 * return updated position in program line */
char* setINPUTarray(char *pos, double value) {
	int M=0, N=0, T=0, var=0;
	char *h=pos;

	/* find closing parenthesis and replace it 
	 * with a space, to stop the parser */
	while (*h!=')') h++; *h=' ';

	/* get var index */
	if (issv(pos)) var=*pos, pos+=2;
	else if (isdv(pos)) var=in(*pos,*(pos+1)), pos+=3;
	else {perror_(6,l,false_); return NULL;}

	/* read array element coordinates */
	p=pos, M=S(), T=1;
	/* has more than one dimension? */
	if(*p==',') p++, T=2, N=S();
	setArrayVal(var,T,M,N,value);
	return ++p;
}


/* The following routines define the behavior of standard (and extended)
 * BASIC statements, as explained in the 1964 version of the manual;
 * the INPUT routine behavior, present in version III and absent in version
 * II and in the 1964 manual, is mostly taken from the 1975 I edition of  
 * "Programming with BASIC" by Byron S. Gottfried (Italian version). */ 


/* DIM()
 * The DIM statement */
void DIM(void){
	char *pp=p-1;
	
	do {
		int var=0, M=0, N=0, T=1;
		/* take var index */
		pp++;
		if (issv(pp)) var=*pp++;
		else if (isdv(pp) && ef) var=in(*pp,*(pp+1)), pp+=2;
		else {perror_(6,l,false_); return;}

		/* get M (and N) */
		if (*pp!='(') {perror_(7,l,false_); return;}
		pp++; // skip '('
		p=pp; M=(int)S(); pp=p;
		if (M<0) perror_(20,l,true_);
		if (M==0 && ob) {
			perrorEx_(35,0);
			perror_(20,l,true_);
		}
		if(*pp==',') {
			pp++; // skip ','
			p=pp; N=(int)S(); pp=p;
			if (N<0) perror_(20,l,true_);
			if (N==0 && ob) {
				perrorEx_(35,0);
				perror_(20,l,true_);
			}
			T=2;
		}
		/* if option base is 0 and both M and N are zero
		 * (using for instance DIM A(0) or DIM A(0,0) for
		 * respectively a vector with one element and a matrix
		 * with one element) the algorithm would set up a 11-elements
		 * vector or a 11x11 elements matrix; so, I must catch this
		 * fact to warn the programmer */
		if (!ob && !M && !N) perrorEx_(37,l);
		/*  check syntax */
		if (*pp++!=')') {perror_(7,l,false_); return;}
		dimArray(var,T,M,N);
		if (*(p+1)!=','&&*(p+1)!='\0'&&*(p+1)!='\n') {
			perror_(7,l,false_);return;}
	}
	while (*pp==','); 
}


/* INPUT()
 * The INPUT statement */
void INPUT(void) {
	char *pp; // pointer to program line
	char line[STRLIM+1], linens[STRLIM+1];  // user input
	char *dp=line, *dpns=linens; // pointer to user input
	int var=0;
	int retain=FALSE; // flag for void input of one-var

	/* check if last character is something else than comma 
	 * or semicolon, in which case it's an error */
	pp=p;
	while (*pp) pp++;
	if (iscs(pp)) { perror_(5,l,false_); return; }
	
	/* enable string printing as an extended feature;
	 * semicolon, comma or nothing may separate the string 
	 * from the variable list */
	if (*p=='\"' && !ef) {perror_(6,l,false_); return;}
	else if (*p=='\"') {
		p++;
		while (*p!='\"') putchar(*p++);
		p++;
		if (iscs(p)) p++;
	}

	/* 'retain' is put to TRUE if there's only one variable to input;
	 * in this case the user is free to type ENTER on a void line,
	 * thus retaining the original variable value. */
	if (!strstr(p,",") && !strstr(p,";")) retain=TRUE;

	pp=p-1;
	printf("? ");
	line[0]='\0';
	/* take user input */
	fgets(line, STRLIM, stdin);

	/* return if the user pressed ENTER on a void line
	 * with a one-variable input */
	if (line[0]=='\n' && retain) {
		printn=counter=0;
		return;
	}
	
	/* process normal input, turning it into a line without spaces */
	line[strlen(line)-1]='\0'; // get rid of '\n'
	stripBlanks(dp,dpns);
	/* repeat for all variables: */
	do {
		pp++;
		/* if not all values have been input
		 * take another user input line ...*/
		while (!*dpns) {
			int i;
			for(i=0;i<STRLIM;i++) line[i]=linens[i]='\0';
			printf("?? ");
			fgets(line, STRLIM, stdin);
			line[strlen(line)-1]='\0'; dp=line; dpns=linens;
			stripBlanks(dp,dpns);
		}
		/* set var value to input array var*/
		if (isav(pp)) {
			double value;
			p=dpns; value=S(); dpns=p;
			pp=setINPUTarray(pp,value);
			if (*pp) {
				if (!iscs(dpns) && *dpns!='\0') 
					{perror_(7,l,false_);return;}
				if (iscs(dpns)) dpns++;
				continue;
			}
			else break;
		}
		/* take var index of single vars */
		else if (issv(pp)) var=*pp++;
		else if (isdv(pp)) var=in(*pp,*(pp+1)), pp+=2;
		else {perror_(6,l,false_); return;}
		/* set var value to input single var */
		p=dpns; P[var]=S(); dpns=p; 
		if (!iscs(dpns) && *dpns!='\0') {perror_(7,l,false_);return;}
		if (iscs(dpns)) dpns++; // there's more in the user input
	}
	while (iscs(pp)); // there's more in the program line

	/* reset printing data */
	printn=counter=0;
}


/* RANDOMIZE()
 * The RANDOMIZE statement 
 * only for extended features - the check is done by RUN() */
void RANDOMIZE(void) {
	time_t now=time(NULL);
	srand((double)now);
	rand();
	rand();
}


/* ONGOTO()
 * The ON..GOTO statement 
 * only for extended features - the check is done by RUN() */
void ONGOTO(void) {
	int addr[35],ac=0;
	int val;
	if (!check(p,"GOTO")) { 
		if (check(p,"THEN"));
		else {perror_(7,l,false_); return;}
	}
	
	val=(int)S();
	/* check for a valid value (must be greater than zero) */
	if (val<=0) perror_(12,l,true_);

	if(!strncmp(p," OTO",4) || !strncmp(p," HEN",4)){
		p+=4;
		/* at least an address must be given */
		if (*p=='\0') {perror_(7,l,false_); return; }
		addr[ac++]=(int)S();
		while (*p++==',') addr[ac++]=S();
		/* ON..GOTO overflow causes a halt 
		 * see T. Kurtz "BASIC Sessions", page 528 */
		if (val>ac) l=-1; 
		else l=addr[val-1]-1;
		return;
	}
	else {perror_(7,l,false_); return; }
}


/* IFTHEN()
 * The IF..THEN/IF..GOTO statements
 * IF..GOTO is available only as an extended feature
 * IF..THEN followed by assignments is available 
 * only as an extended feature */
void IFTHEN(void){
	int flag, isTHEN=TRUE;
	if (!check(p,"THEN")) {
		if (check(p,"GOTO") && ef) isTHEN=FALSE;
		else {perror_(7,l,false_); return;}
	}

	/* get expression result */
	flag=(int)S();
	
	/* admit IF..THEN GOTO only as an extended feature and skip them */
	if(!strncmp(p," HENGOTO",8) && ef) p+=8;
	/* else skip THEN or GOTO */
	else if(!strncmp(p," HEN",4) || !strncmp(p," OTO",4)) p+=4;
	/* else it's an incorrect IF */
	else (perror_(12,l,true_));
	
	/* if IF..THEN condition is true, perform what follows THEN/GOTO */
	if (flag) {
		/* if after THEN/GOTO there's a destination */
		if (isdigit(*p)) GOTO();
		/* or if after THEN there's a valid assignment */
		else if (isTHEN && ef && LET());
		/* else it's an incorrect address */
		else (perror_(12,l,true_));
	}
}


/* GOTO()
 * The GOTO Statement - called also by IF..THEN
 * a jump outside of a FOR cycle is interpreted as a break for that cycle */
void GOTO(void){
        int num=(int)S();

        /* check if destination is legal */
        if (num<=0 || (!m[num])) perror_(12,l,true_);
        else l=num;
	/* if destination is outside of a FOR cycle domain, break */
	while (forTOS>0 && !isGOSUB) {
		if (l<forStack[forTOS-4] || l>forStack[forTOS-5]) {
			forTOS-=6;
			forcount--;
		}
		/* stop break if jump is inside another FOR */
		if (forTOS>0) {
			if (l>forStack[forTOS-4] && l<=forStack[forTOS-5])
				break;
		}
	}
	l--;
}


/* GOSUB()
 * The GOSUB Statement
 * From the Oct. 1964 manual, page 43:
 * "The user must be very careful not to write a program in which a GOSUB 
 * appears inside a subroutine which itself is entered via a GOSUB; it just 
 * won't work."
 * dib standard version, according to the original Dartmouth BASIC, accepts 
 * one level depth only; GOSLIM is set to 80 levels in dib.h, and main() 
 * changes it to 1 if extended features are not requested; 
 * if you want more depth than 80, change it modifying GOSLIM into dib.h, 
 * recompiling and invoking dib with the -x option */
void GOSUB(void){
        int num=(int)S();
        /* check if destination is legal */
        if (num<=0 || !m[num]) perror_(12,l,true_);
        if (gosubTOS >= GOSLIM) {perrorEx_(34,l); return;}
        else gosubStack[++gosubTOS]=l, l=num-1;
	isGOSUB=TRUE;
}


/* RETURN()
 * The RETURN Statement */ 
void RETURN(void){
	/* RETURN without GOSUB? */
        if (gosubTOS <= 0) {perror_(21,l,false_);return;} 
        /* restore pointers */
        l=gosubStack[gosubTOS--];
	if (gosubTOS==0) isGOSUB=FALSE;
}


/* FOR()
 * The FOR..TO..STEP statement */
void FOR(void){
        int var,i,ivar, nextstep=0;
        double start, end, step;

	if(forcount >= FORLIM) perror_(16,l,true_);

        if (issv(p)) var=*p++;
	else if (isdv(p)) var=in(*p,*(p+1)),p+=2;
	else perror_(7,l,false_);

        if (*p++!='=') perror_(7,l,false_);
        
	if (check(p,"TO")) start=S(), P[var]=start;
	else perror_(7,l,false_);

	if (*p != ' ' && *(p+1) !='O') perror_(7,l,false_);
	
	if (check(p,"STEP")) {
		p+=2;
		end=S();
		p+=4;
		if (!*p) perror_(7,l,false_);
		step=S();
	}
	else {
		p+=2;
		if (!*p) perror_(7,l,false_);
		end=S();
		step=1.0;
	}

	/* look for closing NEXT matching the variable */
	for(i=l;i<CHUNKS*LIMIT;i++){
		if(m[i] && (p=strstr(m[i],"NEXT"))){
			p+=4;
			while (*p==' '||*p=='\t') p++;
			if (issv(p)) ivar=*p;
			else if (isdv(p)) ivar=in(*p,*(p+1));
			else ivar=0;
			if (ivar==var) {
				nextstep=i;
				break;
			}
		}
	}
	
	/* If nextstep is still null, no closing NEXT has been found,
	 * so end program emitting a warning */
	if (!nextstep) {
		perror_(18,l,false_);
		l=-1;
		return;
	}
	
	/* if step is meaningless, go beyond FOR..NEXT */
	if ((end-start>0 && step<0) || (end-start<0 && step>0) || step==0) {
		l=nextstep;
		P[var]-=step;
                return;
	}

        /* update stack values */
	forStack[++forTOS]=nextstep;	// TOS-5 = NEXT program line
	forStack[++forTOS]=l;           // TOS-4 = next instruction
        forStack[++forTOS]=var;		// TOS-3 = var index
        forStack[++forTOS]=end;         // TOS-2 = end value
        forStack[++forTOS]=step;        // TOS-1 = step value
        forStack[++forTOS]=start;       // TOS   = start/current value
	forcount++;
}


/* NEXT()
 * The NEXT... Statement
 * it **must** be followed by the variable name */
void NEXT(void) {
        int var;
        double step = forStack[forTOS-1];
        double end  = forStack[forTOS-2];

        /* get index; if different from what stored with the previous
	 * FOR instruction => NEXT error */
        if (issv(p)) var=*p++;
	else if (isdv(p)) var=in(*p,*(p+1)),p+=2;
	else perror_(7,l,false_); // !
	if (var != (int)forStack[forTOS-3]) perror_(17,l,false_);

        /* if beyond, then end cycle */
        if(((step > 0) && (forStack[forTOS]+step > end)) || \
		((step < 0) && (forStack[forTOS]+step < end))) {
                        forTOS -= 6;
			forcount--;
			return;
	}
        /* else reiterate
	 * update counter and set counter variable */
        forStack[forTOS] += step; // current = current +/- step
        P[var]=forStack[forTOS];
        l=(int)forStack[forTOS-4];
}


/* DEFMUL()
 * The DEF Statement with multiple arguments (only for extended mode)
 * The variable labeling FN must be in the A..Z range and not subscripted
 * (storing only - actual command is performed into LOAD) */
void DEFMUL(void) {
	int len, ward=*(p+2);

	if (*p=='F' && *(p+1)=='N' && isalpha(ward) && *(p+3)=='(') {
		int varcount=0;

		/* point to the first variable */
		p+=4;

		/* store all argument-variables; in extended mode any number 
		 * of any variable type A..Z and A0..Z9 may be used */
		while (*p!=')') {
			if (issv(p)) {
				fn[ward].vars[0]=++varcount;
				fn[ward].vars[varcount]=*p;
				p++;
			}
			else if (isdv(p)) {
				fn[ward].vars[0]=++varcount;
				fn[ward].vars[varcount]=in(*p,*(p+1));
				p+=2;
			}
			else {perror_(7,l,false_); return;}
			/* check syntax */
			if (*p==',') p++;
			else if (*p!=')') {perror_(7,l,false_); return;}
		}
		p++;
		if (*p!='=') {perror_(7,l,false_); return;}
		p++;
		len=strlen(p);
		fn[ward].def=malloc(len+1);
		if (!fn[ward].def) derror_("dib: out of memory for DEF.\n");
		strncpy(fn[ward].def,p,STRLIM);
		fn[ward].def[len]='\0';
	}
	else perror_(7,l,false_);
}


/* DEF()
 * The DEF Statement with one argument (only for standard mode)
 * The variable labeling FN must be in the A..Z range and not subscripted
 * (storing only - actual command is performed into LOAD) */
void DEF(void) {
	int len, ward=*(p+2);
	if (*p=='F' && *(p+1)=='N' && isalpha(ward) && *(p+3)=='(') {
		/* as stated at page 41 of the 1964 BASIC manual
		 * "...the variable used in the DEF statement 
		 * must not be subscripted..."; so I check only 
		 * for one variable of type A..Z in default mode */
		if (issv(p+4) && *(p+5)==')') {
			len=strlen(p+7);
			fn[ward].def=malloc(len+1);
			if (!fn[ward].def)
				derror_("dib: out of memory for DEF.\n");
			strncpy(fn[ward].def,p+7,STRLIM);
			fn[ward].def[len]='\0';
			fn[ward].vars[0]=1;
			fn[ward].vars[1]=*(p+4);
			
		}
		else perror_(7,l,false_);
	}
	else perror_(7,l,false_);
}


/* READ()
 * The READ statement */
void READ(void) {
	/* there must be at least one DATA to read! */
	if (dataLimit==0) {
		perror_(10,l,false_);
		return;
	}
	do {
		if (dc>dataLimit) {
			/* if in extended mode, print a warning */
			if (ef) perrorEx_(22,l);
			/* now exit; cfr. 1964 Dartmouth BASIC manual, p.27 */
			l=-1;
			forcount=0;
			return;
		}
		if (isav(p)) {
			int var, M=0, N=0, T=1;
			if (issv(p)) var=*p++;
			else if (isdv(p)) var=in(*p,*(p+1)), p+=2;
			else {perror_(6,l,false_); return;}
			if (*p!='(') {perror_(7,l,false_); return;}
			p++; // skip '('
			M=(int)S();
			if (M<0) perror_(20,l,true_);
			if(*p==',') {
				p++; // skip ','
				N=(int)S();
				if (N<0) perror_(20,l,true_);
				T=2;
			}
			if (*p!=')') {perror_(7,l,false_); return;}
			setArrayVal(var,T,M,N,D[dc++]);
			p++;
		}
		else if (issv(p)) { 
			P[(int)*p++]=D[dc++];
		}
		else if (isdv(p)) {
			P[in(*p,*(p+1))]=D[dc++], p+=2;
		}
	} while (*p++==',');
}


/* RESTORE
 * the RESTORE statement 
 * if used without DATA statement, emit a NO DATA warning only in case of
 * extended mode; this is the only case in which an error 0-21 is used as
 * a warning */
void RESTORE(void){
	if (dataLimit) dc=1;
	else {
		if (ef) perrorEx_(10,l);
		dc=0;
	}
}


/* PRINT()
 * The PRINT statement */
void PRINT(void) {
	/* PRINT without argument */
	if(!*p) printn=1;
	/* PRINT with argument */
	else while (*p) {
		/* print a string */
		if (*p=='\"') {
			p++;
			while (*p!='\"' && *p!='\0') {
				putchar(*p++);
				counter++;
				if (counter>SCRLIM) {
					CR(); 
					printn=0;
					return;
				}
			}
			/* skip closing quotes */
			p++;
			printn=0;
			/* if after a comma/semicolon there's a TAB,
			 * comma/semicolon function must not be performed */
			if (!strncmp(p,";TAB(",5) || !strncmp(p,",TAB(",5)) {
					p++;
					continue;
			}
			/* as stated at pages 31-32-34, an expression
			 * following the closing double quote prints the 
			 * expression together with the preceding string; 
			 * To this subject, I must notice that the example at
			 * page 32 (the first one) and the one at page 37
			 * of the October 1964 manual show that "together" 
			 * means that the number is anyway printed at the 
			 * beginning of the next 3-chars zone.
			 * This is a different output than that on the DTSS
			 * simulator by Kurtz. */
			if (!iscs(p) && *p!='\0') {
				printn=0;
				*--p=';'; // force semicolon output
				continue;
			}
			printn=1;
			isTab=FALSE;
			continue;
		}
		/* CHR$() function 
		 * Only for extended features */
		else if (!strncmp(p,"CHR$(",5) && ef) {
			char c;
			p+=5;
			c=(char)S();
			if (')'==*p) p++; // skip ')'
			else {perror_(2,l,false_); return;}
			printf("%c",c);
			if (!*p) printn=1; 
			else if (iscs(p) && !*(p+1)) p++,printn=0;
			/* if there's another CHR$, don't do anything */
			else if (iscs(p) && !strncmp(p+1,"CHR$(",5)) {
				p++;
				continue;
			}
			continue;
		}
		/* adjust comma-tab position */
		else if (*p==',') {
			if (!counter) AC();
			while (counter%15) AC();
			p++;
			if (counter>SCRLIM) CR();
			printn=0;
			isTab=FALSE;
		}
		/* adjust semicolon-tab position */
		else if (*p==';') {
			if (!counter) AC();
			while (counter%3) AC();
			p++;
			if (counter>SCRLIM) CR();
			printn=0;
			isTab=FALSE;
		}
		/* TAB() function
		 * As stated at page 528 of the Kurtz essay "BASIC sessions",
		 * to the IV version from 1967 "...a TAB function was added, 
		 * with the column positions numbered from 0 to 74..." */
		else if (!strncmp(p,"TAB(",4) && ef) {
			int i,tab;
			p+=4;
			tab=(int)S();
			if (')'==*p) p++; // skip ')'
			else {perror_(2,l,false_); return;}
			if (tab < 0) {perror_(1,l,false_); return;}
			else if (tab <= SCRLIM) {
				if (tab > counter)
					for (i=0; i<tab-counter; i++)
						putchar(' ');
				counter=tab;
				if (counter>SCRLIM) { CR(); printn=0; return; }
				isTab=TRUE;
				if (iscs(p)) {
					p++;
					continue;
				}
			}
			else { CR(); printn=0; return; }
		}
		/* print an expression (treating also comma and semicolon,
		 * which require a different work with respect to the above
		 * routines designed for strings) */ 
		else {
			int cw;
			cw=counter;
			counter=printNUM(S());
			printn=1;
			/* if after a comma/semicolon there's a TAB,
			 * comma/semicolon function must not be performed */
			if (!strncmp(p,";TAB(",5)||!strncmp(p,",TAB(",5))
				p++;
			/* otherwise perform the comma function... */
			else if (*p==',' && !isTab) {
				int d=counter;
				/* advance to next zone */
				while (counter%15) AC();
				/* if too close, advance to next zone again */ 
				if ((counter-d)<2) {
					AC();
					while (counter%15) AC();
				}
				p++;
				if (counter>SCRLIM-10) CR();
				printn=0;
			}
			/* ... or the semicolon function */
			else if (*p==';'&& !isTab) {
				/* advance to next minizone, with 6 char
				 * minimum, and if too close, go to next */
				int d=6-(counter-cw);
				while (d<2) d+=3;
				counter+=d;
				while (d-- > 0) putchar(' ');
				p++;
				if (counter>SCRLIM-10) CR();
				printn=0;
			}
			/* if next there's a string without separator,
			 * push text to next minizone, and if already at the
			 * beginning of a minizone, skip to next */
			else if (*p=='\"' && !isTab) {
				int skip=FALSE;
				while (counter%3) AC(), skip=TRUE;
				if (!skip) { 
					AC(); while (counter%3) AC();
				       	printn=0;	
				}
			}	
			isTab=FALSE;
		}
	}
	if (printn) CR();
}


/* LOAD(filename-string)
 * the loader */
void LOAD(char *filename) {
	/* the following 'endl' work as a flag for the END statement, 
	 * holding the line number in which it appears,
	 * while 'endn' holds the number of END statements found */
	int i;
	FILE *f;

	/* trying to open file */
	f=fopen(filename,"r");
	if (!f) derror_("dib: file not found.\n");
	
	/* erase memory */
	for(i=0;i<CHUNKS*LIMIT;i++) m[i]=0;

	while(fgets(B,STRLIM+1,f)) {
		int i;
		/* strip away '\n' left by fgets() */
		B[strlen(B)-1]='\0';
		/* strip away ASCII 13 left by DOS files */
		if (B[strlen(B)-1]==13) B[strlen(B)-1]='\0';
		/* store line */
		G(); // in G() p is set equal to B
		while (isspace(*p)) p++;
		/* DATA Statement - saving elements */
		if (!strncmp(p,"DATA",4)) {
			p+=4;
			while (*p) {
				/* DATA values are 0-299 for default mode
				 * and 0-998 for extended mode */
				if((!ef && dc>=DATLIM-1)||(ef && dc>=LIMIT-1)){
                                        perror_(14,l,true_);
                                        break;
                                }
				while (isspace(*p)) p++;
				D[++dc]=S();
				while (isspace(*p)) p++;
				if (*p=='\'') while (*p) p++;
				if (*p!=',' && *p!='\0') {
					perror_(7,l,false_);
					break;
				}
				p++;
			}
		}
		/* DEF Statements - saving functions */
		else if (!strncmp(p,"DEF",3)) {
			char *x=J;
			p+=3;
			stripBlanks(p,x); p=J;
			if (ef) DEFMUL(); 
			else DEF();
		}
		/* END statements - check */	
		else if (!strncmp(p,"END",3)) {
			endn++;
			/* check if END is unique */
			if (endn>1 && !ef) perror_(8,endl,true_);
			endl=l;
		}
		/* clear buffer*/
		for (i=0;i<STRLIM;i++) B[i]='\0';
	} /* end reading and storing file lines */

	/* trying to close file */
	if (fclose(f)) derror_("dib: unable to close file.\n");

	/* set up parameters for READ */
	dataLimit=dc, dc=1;

	/* check if END exists */
	if (!endn && !ef) perror_(9,l,true_);

	/* check if END is not last */
	if (endl!=l && !ef) perror_(8,endl,true_);
}


/* LET()
 * The LET statement (even for implicit LET)
 * assignment status: 
 * - for single assignment, return assignment status unless an error
 *   is met, in this case return FALSE
 * - for multiple assignments, return last assignment status,
 *   or FALSE at the first non-working assignment */
int LET(void) {
	int res=TRUE;
	/* in default mode, only LET statements (with LET required)
	 * and a single assignment may be used*/
	if (!ef) {
		/* if it doesn't begin with LET it's an error */
		if (strncmp(p,"LET",3)) res=FALSE;
		else {
			p+=3;
			res = assign();
			if (*p) res=FALSE;
		}
	}
	/* in extended mode, LET may be omitted and multiple assignments 
	 * (separated by comma or colon) may be used*/
	else {
		int div; // divider may be comma or colon
		if (!strncmp(p,"LET",3)) p+=3;
		do {
			div = FALSE;
			res = assign();
			if (*p==',' || *p==':') div = TRUE;
			p++;
		} while (res && div);
	}

	return res;
}


/* assign()
 * The assign statement - called by LET() 
 * return TRUE if assingment went OK, false otherwise */
int assign(void) {
	char *e=p;
	if (isav(p)) dimAssign();
	else if (issv(p)) p+=2, P[(int)*e]=S();
	else if (isdv(p)) p+=3, P[in(*e,*(e+1))]=S();
	else return FALSE;
	return TRUE;
}


/* RUN()
 * The RUN Statement */
void RUN(int starting_line) {
	int i;
	
	/* set starting line */
	l=starting_line;
	
	/* initialize variables */
	for(i=0; i<LIMIT;i++) P[i++]=0.0;

	/* execute line by line */
	while(l){
		/* sub means 'shall I substitute?' */
		int qq, sub=1;

		/* empty buffer */
		for (qq=0;qq<=STRLIM;qq++) B[qq]='\0';
		
		/* skip unexistent lines in memory */
		while(!(s=m[l])) l++;
	
		/* debug output - print unchanged line */
		if (debug) {
			if (!printn) putchar('\n');
			printf("%d %s\n",l,m[l]);
		}
		
		/* tokenize '#' for <>, '$' for <=, '!' for >=, '&' for AND, 
		 * '|' for OR, '%' for XOR/EOR, '_' for EQV, '~' for IMP, 
		 * '}' for \ (integer division) and '{' for MOD 
		 * (if they are not into a string)
		 * also, put a '\0' where an apostrophe comment is found 
		 * - NOT() is treated like a unary operator in getVal() 
		 * - AND is tokenized if not into RANDOMIZE
		 * - OR is tokenized if not into FOR, NORM or RESTORE */ 
		p=s; d=0;
		while (*p){
			if (*p=='\"') sub = (sub ? 0 : 1);
			if (sub) {
				if (*p=='<' && *(p+1)=='>') *p++='#',*p=' ';
				else if (*p=='<'&& *(p+1)=='=') *p++='$',*p=' ';
				else if (*p=='>'&& *(p+1)=='=') *p++='!',*p=' ';
				else if (!strncmp(p,"AND",3) &&\
					        strncmp(p,"ANDO",4) && ef) 
					*p++='&',*p++=' ',*p=' ';
				else if (!strncmp(p,"OR",2) &&\
						strncmp(p-1,"FOR",3) &&\
						strncmp(p,"ORM(",4) &&\
					       	strncmp(p,"ORE",3) && ef) 
					*p++='|',*p=' ';
				else if (!strncmp(p,"XOR",3) && ef) 
					*p++='%',*p++=' ',*p=' ';
				else if (!strncmp(p,"EOR",3) && ef) 
					*p++='%',*p++=' ',*p=' ';
				else if (!strncmp(p,"EQV",3) && ef) 
					*p++='_',*p++=' ',*p=' ';
				else if (!strncmp(p,"IMP",3) && ef) 
					*p++='~',*p++=' ',*p=' ';
				else if (!strncmp(p,"MOD",3) && ef) 
					*p++='{',*p++=' ',*p=' ';
				else if (*p=='\\' && ef) *p='}';
				else if (*p=='\'') *p='\0';
			}
			p++;
		}
		
		/* strip blanks; afterwards, buffer B will hold
		 * executing line */
		d=B; stripBlanks(s,d);
		
		/* if it's not an assignment, it's a statement,
		 * so process language */
		if(B[1]!='=' && B[2]!='=' && !isaa(B)) {
			     if (B[0]=='\0'); // nop for void lines
			else if (!strncmp(B,"CHANGE",6) && ef) p=B+6, CHANGE(); 
			else if (!strncmp(B,"DATA",4)); //nop
			else if (!strncmp(B,"DEBUG",5)) debug=TRUE; // !! 
			else if (!strncmp(B,"DEF",3)); //nop
			else if (!strncmp(B,"DEG",3) && ef) degree=DEGREE;
			else if (!strncmp(B,"DIM",3)) p=B+3, DIM();
			else if (!strncmp(B,"END",3)) l=-1;
			else if (!strncmp(B,"EXTEND",6)) ef=TRUE;
			else if (!strncmp(B,"FOR",3)) p=B+3,FOR();
			else if (!strncmp(B,"GOSUB",5))	p=B+5,GOSUB();
			else if (!strncmp(B,"GOTO",4)) p=B+4,GOTO();
			else if (!strncmp(B,"IF",2)) p=B+2, IFTHEN();
			else if (!strncmp(B,"INPUT",5))	p=B+5,INPUT();
			else if (!strncmp(p=B,"LET",3) && LET());
			else if (!strncmp(B,"MATINPUT",8)) p=B+8, MATINPUT();
			else if (!strncmp(B,"MATPRINT",8)) p=B+8, MATPRINT();
			else if (!strncmp(B,"MATREAD",7)) p=B+7, MATREAD();
			else if (!strncmp(B,"MAT",3)) p=B+3, MATtype();
			else if (!strncmp(B,"NEXT",4)) p=B+4, NEXT();
			else if (!strncmp(B,"NODEBUG",7)) debug=FALSE; // !!
			else if (!strncmp(B,"ON",2) && ef) p=B+2, ONGOTO();
			else if (!strncmp(B,"OPTIONBASE",10) && ef) {
				p=B+10; ob=((int)S())?1:0; }
			else if (!strncmp(B,"PAGE",4)) system("clear");
			else if (!strncmp(B,"PRINT",5)) p=B+5, PRINT();
  		        else if (!strncmp(B,"RAD",3) && ef) degree=RADIAN;
			else if (!strncmp(B,"RANDOMIZE",9) && ef) RANDOMIZE();
			else if (!strncmp(B,"READ",4)) p=B+4, READ();
			else if (!strncmp(B,"REM",3)); // nop
			else if (!strncmp(B,"RESTORE",7)) p=B+7, RESTORE();
			else if (!strncmp(B,"RETURN",6)) RETURN();
			else if (!strncmp(B,"STOP",4)) l=-1;
			/* all statements are discarded? 
			 * It's an illegal statement*/
			else perror_(5,l,false_);
		}
		/* if it's not a statement, it's an implicit assignment to 
		 * variable, enabled only with extended features */
		else if ((p=B) && LET());
		/* If it's not a statement nor an implicit assignment,
		 * it's an illegal instruction */
		else perror_(5,l,false_);
		l++;
	}
	/* check if some NEXT was forgotten */
	if (forcount) perror_(18,forStack[forTOS-4],false_);

	/* Carriage Return */
	CR();
}


/* G()
 * store program line
 * the algorithm accepts lines as 10PRINT or 10 PRINT, being the statement
 * always PRINT; this is of little importance for non interactive usage, but 
 * in prevision to implement dib as a standalone system, I want to make
 * sure that the user is free to type what he likes; 
 * in the listing, empty lines are accepted, even if they contain blanks
 * lines beginning with zero are ignored, as lines beginning with # 
 * (scripting purposes) 
 * the limit to 99999 as the highest line number is not considered
 * in the extended features
 * This function G() is mainly due to Diomidis Spinellis, and retains
 * its original name, but it has been heavily managed by me */
void G(void){
	p=B;
	if(*p=='\0'||*p=='#') return;
	if(isalpha(*p)) perror_(4,0,true_);
	l=atoi(p);
	if(!ef && l>99999) perror_(4,l,true_);
	if(ef && l>(CHUNKS*LIMIT)){
		char temp[40];
		sprintf(temp, "dib: line %d is beyond domain.\n", l);
		derror_(temp);
	}

	/* Erase line if existing */
	if (m[l]) {int i=0; while (m[l][i]) m[l][i++]='\0'; }
	
	/* skip line number digits */
	while(isdigit(*p)) p++;
	/* skip blanks */
	while(isspace(*p)) p++;
	/* store */
	if (*p!='\0') {
		m[l]=malloc(strlen(p)+1);
		if (!m[l]) {
			char temp[40];
			sprintf(temp, "dib: out of memory for line %d.\n",l);
			derror_(temp);
		}
		strncpy(m[l],p,STRLIM);
	}
} 


/* The following functions constitute the whole parser engine:
 * S() 		: the expression parser
 * getVal()	: the single value parser 
 * getFunc()	: the function parser
 * priority()	: return operator precedence
 * These functions together have been completely rewritten on completely
 * different bases than those constituting the Spinellis version - actually,
 * they are a different thing
 * Copyright Antonio Maschio - April 2008 */


/* S()
 * The PARSER for expressions - an algebraic expression is turned into an 
 * RPN stack-based expression, with numbers and operators in order in their own
 * stack; operations are performed from higher priorities down to lower and 
 * left to right;  priorities are given by the priority() function, which is 
 * part of this tool. The name S() is retained from Spinellis version,
 * due to its wide usage in the code and also as a homage to him. */
double S(void) {
	double numStack[PSTACK];
	int opStack[PSTACK], prStack[PSTACK], numTOS=0, opTOS=0, ii,jj, level;

	/* clear stack buffer */
	//for (ii=0;ii<PSTACK;ii++) numStack[ii]=0;

	/* catch unary plus at start (simply discarded because redundant) */
	if (*p=='+') p++;
	/* catch unary minus at start */
	else if (*p=='-') {
		double res;
		p++;
		res=getVal();
		if (*p=='^') {
			/* power following a unary minus has higher priority */
			double expon;
			int sign=1;
			p++;
			if (*p=='-') p++, sign=-1;
			else if (*p=='+') p++;
			expon=sign*getVal();
			res=-power(res,expon);
		}
		else res=-res;
		numStack[numTOS++]=res;
		opturn=1;
	}
	/* reverse algebraic expression into a stack-based structure */
	while (*p && !iscs(p) && *p!='\"' && *p!='\'' && !isspace(*p) \
			&& *p!=':'&&*p!='?'){
		/* catch an operator */
		if (isop(p) && opturn) {
			opStack[opTOS]=*p;
			prStack[opTOS]=priority(*p);
			opTOS++;
			p++;
			opturn=0;
			/* operator followed by void is an illegal formula */
			if (!*p) {perror_(2,l,false_);}
		}
		/* catch a closed parenthesis */
		else if (*p==')') break;
		/* catch the unary minus after an operator */
		else if (*p=='-' && !opturn) {
			p++;
			numStack[numTOS++]=-getVal();
			opturn=1;
		}
		/* catch the unary plus after an operator */
		else if (*p=='+' && !opturn) {
			p++;
			numStack[numTOS++]=getVal();
			opturn=1;
		}
		/* catch a number (literal or variable) */
		else {
			numStack[numTOS++]=getVal();
			opturn=1;
		}
	}

	/* perform calculations from higher priorities downward */
	for (ii=5;ii>=0;ii--) {
		/* calculations are performed from left to right */
		for (level=0;level<opTOS;level++) {
			if (prStack[level]==ii) {
				/* collapse stacks */
				switch(opStack[level]) {
/************** HERE BEGINS OPERATORS ANALYSIS *********************/
/* lesser than < */
case '<': numStack[level]=-(double)(numStack[level]<numStack[level+1]); break;
/* greater than > */
case '>': numStack[level]=-(double)(numStack[level]>numStack[level+1]); break;
/* not equal <> */
case '#': numStack[level]=-(double)(numStack[level]!=numStack[level+1]); break;
/* greater than or equal >= */
case '!': numStack[level]=-(double)(numStack[level]>=numStack[level+1]); break;
/* lesser than or equal <= */
case '$': numStack[level]=-(double)(numStack[level]<=numStack[level+1]); break;
/* lesser than */
case '=': numStack[level]=-(double)(numStack[level]==numStack[level+1]); break;
/* power ^ */
case '^': numStack[level]=power(numStack[level],numStack[level+1]); break;
/* plus + */
case '+': numStack[level]+=numStack[level+1]; break;
/* minus - */
case '-': numStack[level]-=numStack[level+1]; break;
/* times * */
case '*': numStack[level]*=numStack[level+1]; break;
/* normal division */
case '/': {
		double div=numStack[level+1];
		/* if divisor (div) is null, dividend remains as result */
		if (div==0) perrorEx_(27,l); 
		else numStack[level]/=div;  
	  }
	  break;
/* AND */
case '&': numStack[level]=-(double)((int)numStack[level]&&(int)numStack[level+1]); break; 
/* OR */
case '|': numStack[level]=-(double)((int)numStack[level]||(int)numStack[level+1]); break; 
/* XOR / EOR */
case '%': {
		int a=(int)numStack[level];
		int b=(int)numStack[level+1];
		numStack[level]=-(double)(b?!a:!!a);
	}
	break;
/* IMP */
case '~': {
		int a=(int)numStack[level];
		int b=(int)numStack[level+1];
		numStack[level]=-(double)(!(a&&!b));
	}
	break;
/* EQV */
case '_': {
		int a=(int)numStack[level];
		int b=(int)numStack[level+1];
		numStack[level]=-(double)(b?!!a:!a);
	}
	break;
/* MOD */
case '{': { 
		int a=(int)numStack[level];
		int b=(int)numStack[level+1];
		/* if divisor is null or greater in absolute value 
		 * than dividend, dividend remains as result */
		if (b==0) perrorEx_(27,l);
        	else if (abs(b) <= abs(a)) {
			/* perform the Crenshaw theory about modulus */
			int d = (int) a/b;
			if (d<0) d--;
			numStack[level]=a-b*d;
		}
	}
	break;
/* integer division */
case '}': { 
		int a=(int)numStack[level];
		int b=(int)numStack[level+1];
		/* if divisor is null, dividend remains as result */
		if (b==0) perrorEx_(27,l);
        	else {
			int c=(int)a/b;
	        	int k=(a-b*c!=0?1:0);
		        numStack[level]=((a<0)?((c<=0)?c-k:c+k):c);
		}
	}
	break; 
/************** HERE ENDS OPERATORS ANALYSIS *********************/
				}
				/* shift stack positions */
				for (jj=level;jj<opTOS-1;jj++) {
					numStack[jj+1]=numStack[jj+2];
					opStack[jj]=opStack[jj+1];
					prStack[jj]=prStack[jj+1];
				}
				opTOS--;
				level--;
			}
		}
	}
	/* At this point, we have the final result into [0] */
	return numStack[0];
}


/* getVal()
 * get single value (number, variable, function or array element) 
 * the case of the unary minus has been caught in S() */
double getVal(void){
	double o=0;
	/* case of a real number */
	if (isnm(p)) {
		char *basep=p;
		o=strtod(p,&p);

		/* error in number conversion */
		if (basep == p) {
			perror_(1,l,false_);
	        }	
	}
	/* case of unary minus 
	 * (for general cases, e.g. before a parenthesis) */
	else if (*p=='-') {
		p++;
		o=-getVal();
	}
	/* case of unary operator NOT */
	else if (!strncmp(p,"NOT",3) && ef) {
		p+=3;
		o=!getVal();
	}
	/* case of open parenthesis */ 
	else if (*p=='(') { 
		p++;
		o=S();
		if (*(p)!=')') perror_(2,l,false_);
		else p++; 
	}
	/* case of array variable */
	else if (isav(p)) {
		int var=0,M=0,N=0,T=1;
		if (issv(p)) var=*p++;
		else if (isdv(p)) var=in(*(p),*(p+1)),p+=2;
		else perror_(2,l,false_);
		p++; // skip '('
		M=(int)S();
		if (M<0) perror_(20,l,true_);
		if (M==0 && ob) perror_(20,l,true_);
		if(*p==',') {
			p++;
		        N=(int)S();
			if (N<0) perror_(20,l,true_);
			if (N==0 && ob) perror_(20,l,true_);
			T=2;
		}
		o=getArrayVal(var,T,M,N); p++; 
	}
	/* case of variable A..Z */
	else if (issv(p)) { o=P[(int)*p++]; }
	/* case of variable A0..Z9 */
	else if (isdv(p)) { o=P[in(*(p),*(p+1))]; p+=2; }
	/* delegate function calculations to getFunc() */
	else if (isfn(p)) { o=getFunc(), p++; } 
	else perror_(2,l,false_);
	return o;
}


/* getFunc()
 * functions calculator */
double getFunc(void){
	double o;
	     if(!strncmp(p,"ABS(",4)) p+=4,o=fabs(S());
	else if(!strncmp(p,"ATN(",4)) p+=4,o=dg(atan(S()));
	else if(!strncmp(p,"COS(",4)) p+=4,o=cos(rd(S()));
	/* DET() is an extended feature, but I catch it here because
	 * of the following DET (default) call */
	else if(ef && !strncmp(p,"DET(",4)) {
		p+=4;
		MATDET();
		o=detValue;
	}
	else if(!strncmp(p,"DET",3)) {
		if (!isdetCalc) perrorEx_(36,l);
		o=detValue;p+=2;*p=')';
	}
	else if(!strncmp(p,"EXP(",4)) p+=4,o=exp(S());
	/* FNORM is an extended feature, but I catch it here because
	 * of the following FN (default) call */
	else if(ef && !strncmp(p,"FNORM(",6)) o=MATNORM();
	/* DEF FN */
	else if(!strncmp(p,"FN",2)) {
		int ward;
		if (isalpha((ward=*(p+2))) && *(p+3)=='(') {
			int cnt=1;
			char *pp;
			p+=4;
			while (cnt <= fn[ward].vars[0]) {
				P[fn[ward].vars[cnt]]=S();
				p++; // skip comma
				cnt++;
			}
			p--;
			/* save pointer, because p must point 
			 * to stored FN function copied to buffer */
			pp=p; 
			if (*p==')') {
				char buf[STRLIM+1];
				strncpy(buf, fn[ward].def, STRLIM);
				p=buf;
				if (*p) o=S();
				else perror_(7,l,false_);
			}
			else perror_(2,l,false_);
			p=pp; // restore pointer 
		}
		else perror_(11,l,false_);
	}
	else if(!strncmp(p,"INT(",4)) p+=4,o=(int)(S());
	/* the absolute value of LOG was in the original BASIC */ 
	else if(!strncmp(p,"LOG(",4)) {
		p+=4; o=S();
		if (o<0)  {perrorEx_(24,l); o=log(fabs(o));}
		else if (o==0) {perrorEx_(25,l); o=MINFF;}
		else o=log(o);
	}
	else if(!strncmp(p,"NUM",3)) o=numValue, p+=2, *p=')';
	/*  The RND with required parameter (default mode) */
        else if(!strncmp(p,"RND(",4)) {p+=4; S(); o=getRandom(); }
	/*  The RND without required parameter was allowed since version IV
	 *  dib allows its usage as an extended feature */
        else if(!strncmp(p,"RND",3) && ef) { o=getRandom(); p+=2; *p=')'; }
	else if(!strncmp(p,"SGN(",4)) {p+=4, o=S(); if (o) o=(o/fabs(o)); }
	/* the absolute value of SQR was in the original BASIC */ 
	else if(!strncmp(p,"SQR(",4)) {
		p+=4; o=S();
		if (o<0) perrorEx_(23,l);
		o=sqrt(fabs(o));
	}			
	else if(!strncmp(p,"SIN(",4)) p+=4,o=sin(rd(S()));
	else if(!strncmp(p,"TAN(",4)) p+=4,o=tan(rd(S()));
	/* extra math functions associated to extended features:
	 * ASN, ASNH, ACS, ACSH, ATNH, COSH, COT
	 * SINH, TANH, DOT, PI, CNORM, RNORM, FIX, FRAC */
	else if(ef && !strncmp(p,"ASN(",4)) {
		p+=4,o=S();
		if (o<-1 || o>1) {perrorEx_(38,l);o=0;}
		else o=dg(asin(o));
	}
	else if(ef && !strncmp(p,"ASC(",4)) {
		p+=4;
		if (*p=='\"') {
			o=(double)((int)*++p);
			while (*p && *p!='\"') p++;
			p++;
		}
		else {
			o=(double)((int)*p);
			while (*p && *p!=')') p++;
		}
	}
	else if(ef && !strncmp(p,"ASNH(",5)) p+=5,o=asinh(S());
	else if(ef && !strncmp(p,"ACS(",4)) {
		p+=4,o=S();
		if (o<-1 || o>1) {perrorEx_(38,l);o=0;}
		else o=dg(acos(o));
	}
	else if(ef && !strncmp(p,"ACSH(",5)) {
		p+=5,o=S();
		if (o<1) {perrorEx_(39,l);o=0;}
		else o=acosh(o);
	}
	else if(ef && !strncmp(p,"ATNH(",5)) p+=5,o=atanh(S());
	else if(ef && !strncmp(p,"COSH(",5)) p+=5,o=cosh(S());
	else if(ef && !strncmp(p,"COT(",4)) {
		p+=4,o=S();
		if (o==0) {perrorEx_(40,l);o=INFF;}
		else o=1.0/tan(rd(o));
	}
	else if(ef && !strncmp(p,"CNORM(",6)) o=MATNORM();
        else if(ef && !strncmp(p,"DOT(",4))  p+=4,o=MATDOT();
        else if(ef && !strncmp(p,"FIX(",4))  p+=4,o=FIX();
	else if(ef && !strncmp(p,"FRAC(",5)) p+=5,o=S(),o-=(int)o;
	else if(ef && !strncmp(p,"GAMMA(",6)) p+=6,o=GAMMA();
        else if(ef && !strncmp(p,"PI",2)) p++, *p=')', o=acos(-1.0);
	else if(ef && !strncmp(p,"RNORM(",6)) o=MATNORM();
	else if(ef && !strncmp(p,"SINH(",5)) p+=5,o=sinh(S());
	else if(ef && !strncmp(p,"TANH(",5)) p+=5,o=tanh(S());
	else perror_(2,l,false_);
	/* check closing parenthesis (advancing of p is done into getVal() )*/
	if (*p!=')') perror_(2,l,false_);
	return o;
}


/* priority(operator-char)
 * return math priority for operators, according to the following table
 * (operators marked with ** belong to the extended features)
 * 7: (..) (caught in getVal) have the higher precedence in absolute
 * 6: NOT** (caught in getVal) has the higher precedence among operators
 * 5: power
 * 4: multiplication, division, integer-division** and MOD**
 * 3: addition, subtraction
 * 2: <, >, =, <=, >=, <>
 * 1: AND**, OR**, XOR**
 * 0: EQV**, IMP**
 * */
int priority(const int op) {
	if (op=='^') return 5;
	else if (op=='*'||op=='/'||op=='}'||op=='{') return 4;
	else if (op=='+'||op=='-') return 3;
	else if (op=='<'||op=='>'||op=='='||op=='$'||op=='#'||op=='!')return 2;
	else if (op=='&'||op=='|'||op=='%') return 1;
	else return 0; // EQV and IMP
}


/* The following functions define system features of dib: 
 * number output, printer management, variable maintenance, errors printing, 
 * math operations, licensing, versioning, help and others */


/* AC()
 * Advance Carriage */
void AC(void) {
	counter++;
	putchar(' ');
}


/* CR()
 * Carriage Return */
void CR(void) {
	counter=0;
	putchar('\n');
}


/* getRandom()
 * return a random number */
double getRandom(void) {
	double o=(double)rand();
	while (o>RANDL) o/=16;
	return o/RANDL;
}


/* doExp(str, &base, &exp) 
 * find base and exponent of a number which must be printed in exponential 
 * form given the pointer to the string str containing the number in raw form.
 * Note: the DTSS emulator by T. Kurtz prints exponential numbers in the form:
 * X.XXXXX E YY if the number has a positive exponent, and in the form: 
 * X.XXXXX E-YY if the number has a negative exponent; in the 1964 manual there
 * are no printed examples of such an output, but at page 34, rule 3, the 
 * typewriter text reports a blank before the E, and a minus sign after 
 * (suggesting that a positive exponent would print a space as normal),
 * and thus confirming the DTSS emulator output. */
void doExp(char *pos, double *base, int *exp) {
	char *e = strstr(pos, "E");
	if (!e) return; // make error?
	*e=' ';
	*base=strtod(pos,NULL);
	*exp=strtol(e,NULL,10);
}


/* emitNUM(sign-char,number-string)
 * emit the string number calculated by printNUM(), checking screen limit;
 * - minus holds the character '-' in case of negative number or ' ' in case
 *   of positive number (sign has always its position in Dartmouth BASIC)
 * - num holds the string of the converted number to be printed 
 * return updated position of counter (which is passed through by printNUM) */
int emitNUM(char minus, char *num) {
	int ret=strlen(num)+1;
	if ((counter+ret) > SCRLIM) CR();
	printf("%c%s", minus, num);
	return counter+ret;
}


/* printNUM(number)
 * print number according to Dartmouth BASIC rules at page 34 (1964 manual)
 * return updated position of counter (as yielded by emitNum) */
int printNUM(double num){
	int isFloat=TRUE;
	char minus=' ';
	char nums[STRLIM+1], *nump;
	double base;
	int exp;

	/* if number is null, shorten the process! */
	if (num==0) return emitNUM(' ',"0");
	
	/* operate only with positive numbers, storing the sign aside */
	if (num<0) num=-num , minus='-';

	/* option -z activates cutOff, which causes numbers in output
	 * to be printed as zero if lower than EPSILON (1E-10 by default)
	 * these "fake" numbers have a negative or a positive sign, and
	 * so they can be distinguished by "real" zeroes */
	if (cutOff && num<EPSILON) {
		if (minus=='-') return emitNUM(minus,"0");
		else return emitNUM('+',"0");
	}

	/* (checked against the dtss simulator) number should actually never 
	 * be lesser than ZERO; the retaining of the sign is a dib feature. */
	if (num<ZERO) {
		perrorEx_(31,l);
		return emitNUM(minus,"0"); 
	}

	/* (checked against the dtss simulator)
	 * the number should never be > INFF in absolute value */
	else if (num > INFF) {
		perrorEx_(30,l);
		num=INFF;
	}

	/* use exponential or normal form in representing the number 
	 * in a raw form (for the moment being) */
	if (num < 1E-10 || num > 1E10) sprintf(nums, "%.5E", num);
	else sprintf(nums, "%21.10F", num);

	/* rule 2: suppress final zeroes */
	nump=nums+strlen(nums)-1;
	while('0'==*nump) *nump--='\0';

	/* rule 4 part 1: if there's a final dot, suppress it 
	 * and clear float flag, because the number is an integer */
	if('.'==*nump){
		*nump--='\0';
		isFloat=FALSE;
	}

	/* suppress any possible trailing nulls */
	nump=nums;
	while(*nump == '0' || *nump == ' ') nump++;

	/* print real number */
	if(isFloat){
		/* exception to rule 3: print real number as-is if it fits */
		if (num <= 0.1 && strlen(nump) < 8) return emitNUM(minus,nump);

		/* rule 3: print real with exponent */
		else if (num < 0.1) {
			sprintf(nums,"%.5E",num);
			doExp(nums,&base,&exp);
			if (exp<0) sprintf(nums,"%.5f E%d",base,exp);
			else sprintf(nums,"%.5f E %d",base,exp);
			return emitNUM(minus,nums);
		}
		/* rule 1: print real with 6 significant digits 
		 * retaining decimal dot (use of #)*/ 
		else {
			sprintf(nums,"%#.6G",num);
			nump=strstr(nums,"E");
			/* clean ending trailing zeroes if number is
			 * not in exponential form */
			if (!nump) {
				nump=nums+strlen(nums)-1;
				while (*nump=='0') *nump--='\0';
			}
			/* clean number if in exponential form */
			else {
				doExp(nums,&base,&exp);
				if (exp<0) sprintf(nums,"%.5f E%d",base,exp);
				else sprintf(nums,"%.5f E %d",base,exp);
			}
			/* skip leading trailing zero, e.g. 0.nnn -> .nnn */
			nump=nums;
			if(*nump=='0') nump++;
			return emitNUM(minus,nump);
		}
	}
	/* print integer number */
	else {
		/* rule 4 part 2: print integer till 9 significant digits */
		if (strlen(nump) < 10) return emitNUM(minus,nump);

		/* rule 4 part 2 (reverse case): print integer with exponent
		 * and 6 significant digits (that is, 5 decimals) */
		else { 
			sprintf(nums,"%.5E",num);
			nump=strstr(nums,"E");
			if (nump) {
				doExp(nums,&base,&exp);
				if (exp <0) sprintf(nums,"%.5f E%d",base,exp);
				else sprintf(nums,"%.5f E %d",base,exp);
			}
			return emitNUM(minus,nums);
		}
	}

	/* avoiding warnings.... */
	return 0;
}


/* check(string1,string2)
 * check if sb syntax is correct against sc; if it is, put a space in the 
 * position found, so that a subsequent parsing by S() may stop without 
 * issuing an error message from getFunc() and return true; 
 * if syntax it's not correct, return false.
 * This is done in statements with some core-words, like IF <Test> THEN
 * where the parsing of <Test> must stop just before THEN; check() inserts 
 * a space in place of the T of THEN, causing the parser to stop there */
int check(char *sb, char *sc){
	char *qq=strstr(sb,sc);
	/* put a space to stop S() parsing */
	if (qq) { *qq=' '; return TRUE; }
	return FALSE;
}


/* power(base,exponent)
 * powering function */
double power(double base, double expon) {
	/* if base is negative, take some cautions... */
	if (base<0) {
		/* integer exponent is always evaluable */
		if (expon==(double)((int)expon)) {
			if ((int)expon%2) 
				// odd exponent
				return -pow(-base,expon);
				// even exponent
				else return pow(base,expon);
		}
		/* while real exponent is not if base < 0,
		 * so take absolute value */
		else {
			perrorEx_(26,l);
			/* the absolute value was in the original BASIC */
			return pow(fabs(base),expon);
		}
	}
	/* INFF is a safe result for 0^(<0) */
	else if (expon<0 && base==0) {
		perrorEx_(32,l);
		return INFF;
	}
	/* 1 is the limit of function x^x if x approaches zero from right */
	else if (expon==0 && base==0) {
		perrorEx_(33,l);
		return 1.0;
	}
	/* default case for base > 0 and any exponent */
	else return pow(base,expon);
}


/* dg(rad)
 * convert radians to degrees if not in RADIAN mode */
double dg(double val) {
	if (degree==RADIAN) return val;
	else return 180*val/acos(-1); // acos(-1) is PI
}


/* rd(deg)
 * convert degrees to radians if not in RADIAN mode */
double rd(double val) {
	if (degree==RADIAN) return val;
	else return val*acos(-1)/180; // acos(-1) is PI
}


/* stripBlanks(src, dest)
 * strip blanks from src and copy stripped string to dest
 * stop stripping while into a string */
void stripBlanks(char *src, char *dest) {
	int j=1; // strip flag/counter
	while (*src) {
		if (*src=='\"') j++;
		if (j&1) { if (*src!=' ' && *src!='\t') *dest++=*src; }
		else *dest++=*src;
		src++;
	}
	*dest='\0';
}


/* derror_(error-string)
 * print dib system error message */
void derror_(char* text) {
	fprintf(stderr, text);
	fprintf(stderr, "Type dib -h for help.\n");
	exit(EXIT_FAILURE);
}


/* perrorEx_(error-code,line-number)
 * print extended BASIC error message without exiting 
 * contextually, retain counter position and 
 * update printing position the new line*/ 
void perrorEx_(int error, int line){
	int diff=0;
	if (counter!=0) putchar('\n');
	if (line) printf("%s  IN %d\n", ers[error], line);
	else printf("%s\n", ers[error]);
	/* rebuild position */
	while (diff++ < counter) putchar(' ');
}


/* perror_(error-code,line-number,exit-flag)
 * print BASIC error message and exit if exit-flag go_away is TRUE
 * also, retain counter position and update printing position on the new line*/ 
void perror_(int error, int line, int go_away) {
	int diff=0;
	if (counter!=0) putchar('\n');
	if (line) printf("%-21sIN %d\n", ers[error], line);
	else printf("%s\n", ers[error]);
	/* exit emitting error code as exit status
	 * the exit status is increased by 100, so subtract 100 to get
	 * the error condition. Just in case you need it... */
	if (go_away) exit(error+100);
	/* rebuild position */
	else while (diff < counter) putchar(' '),diff++;
}


/* in(letter-index,number-index)
 * return index for vars A0..Z9 using a math algorithm 
 * to ensure their memory space is not the same of 
 * variables A..Z, to avoid overwriting */
int in(char c, char d) {
	return (200+((int)c-'A')*10+(int)d);
}


/* printHelp()
 * help message */
void printHelp(void) {
	printf("Usage: dib [options] file\n");
	printf("Options:\n");
	printf("  -b                Set arrays base index to 1 (default is 0)\n");
	printf("  -c                %s execution in case of non-blocking errors\n",false_!=FALSE?"Continue":"Stop");
	printf("      --conditions  Print redistribution conditions from GPL3\n");
	printf("  -d                Activate debug mode\n");
	printf("  -h, --help        Display this information and exit\n");
	printf("  -l                Display errors and warnings lists and exit\n");
	printf("  -t                Print a time report after execution\n");
	printf("  -v, --version     Display version and exit\n");
	printf("      --warranty    Print warranty conditions from GPL3\n");
	printf("  -x                Toggle extended BASIC features setting (default is %s)\n",ef==0?"off":"on");
	printf("  -z                Print as zero numbers smaller than %G\n",EPSILON);
	printf("\nIf no file name is specified, dib will complain.\n");
    	printf("\nYou should have received a copy of the GNU General Public License\n");
	printf("along with this program. If not, see <http://www.gnu.org/licenses/>.\n\n");
	printf("Email bugs reports to <tbin at libero dot it>.\n\n"); 
}


/* printVersion()
 * version message */
void printVersion(int flag) {
printf("\ndib (dib is BASIC) - Dartmouth BASIC III (1966) interpreter\n");
printf("version %s (build %d, %s, %s)\n", VERSION, BUILD, __DATE__, __TIME__);
printf("Copyright (C) %s %s <%s>. ", CY, AU, EM);
if (flag) printf("This program\ncomes with ABSOLUTELY NO WARRANTY; for details type 'dib --warranty'.\n");
if (flag) printf("This is free software, and you are welcome to redistribute it under certain\n");
if (flag) printf("conditions; type 'dib --conditions' for details.");
printf("\nThis software is released under the terms of the GNU General Public License 3.\n\n");
}


/* printWarranty()
 * warranty message (required by GPL3) */
void printWarranty(int flag) {
if (flag) printf("\ndib (dib is BASIC)\nDartmouth BASIC III (1966) interpreter");
if (flag) printf(" - version %s (build %d)\n", VERSION, BUILD);
if (flag) printf("Copyright (C) %s %s <%s>\nunder the terms of the GNU General Public License 3.\n\n", CY, AU, EM);
printf("THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY\n");
printf("APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT\n");
printf("HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM \"AS IS\" WITHOUT\n");
printf("WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT\n");
printf("LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A\n");
printf("PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF\n");
printf("THE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME\n");
printf("THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.\n\n");
}


/* printConditions()
 * conditions message (required by GPL3) */
void printConditions(int flag) {
if (flag) printf("\ndib (dib is BASIC)\nDartmouth BASIC III (1966) interpreter");
if (flag) printf(" - version %s (build %d)\n", VERSION, BUILD);
if (flag) printf("Copyright (C) %s %s <%s>\nunder the terms of the GNU General Public License 3.\n\n", CY, AU, EM);
printf("You may make, run and propagate covered works that you do not convey,\n");
printf("without conditions so long as your license otherwise remains in force.\n");
printf("You may convey covered works to others for the sole purpose of having\n");
printf("them make modifications exclusively for you, or provide you with\n");
printf("facilities for running those works, provided that you comply with the\n");
printf("terms of this License in conveying all material for which you do not\n");
printf("control copyright. Those thus making or running the covered works for\n");
printf("you must do so exclusively on your behalf, under your direction and\n");
printf("control, on terms that prohibit them from making any copies of your\n");
printf("copyrighted material outside their relationship with you.\n\n");
}


/********************************
 *	 dib  MAIN	 *
 ********************************/
int main(int argc, char **argv) {
	int c, ac = argc;
	char *FileName = NULL;
        struct timeval tv;          /* time values        */
        double beginS, endS;        /* reports in seconds */
        int isTiming = FALSE;       /* timer flag         */

	/* parse interpreter arguments */
	while (--ac > 0 && '-' == (*++argv)[0]) {
		/* parse double-dash options */
		     if (!strcmp(*argv, "--help")) 
			{printHelp(); exit(0);}
		else if (!strcmp(*argv, "--version")) 
			{printVersion(TRUE);exit(0);}
		else if (!strcmp(*argv, "--warranty")) 
			{printWarranty(TRUE);exit(0);}
		else if (!strcmp(*argv, "--conditions")) 
			{printConditions(TRUE);exit(0);}
		/* this feature is intentionally undocumented 
		 * I use it to generate the README file */
		else if (!strcmp(*argv, "--greetings")) {
			printf("Welcome, dib user! I hope you\'ll like ");
			printf("this program.\n");
			printf("Here\'s some information about it ");
			printf("and its license...\n");
			printVersion(FALSE);
			printHelp();
			printf("To compile type 'make' as user in ");
			printf("dib source directory.\nTo install ");
			printf("type 'make install' as root after compiling.\n");
			printf("For help about dib's features, after");
			printf(" installation, see 'dib_manual.txt'.\n\n");
			printWarranty(FALSE);
			printConditions(FALSE);
			printf("It's GPL. Enjoy!\n");
			exit(EXIT_SUCCESS);
		}
	
		/* parse single-dash options */
		while ((c = *++argv[0])) {
			     if (c=='b') ob=1; 
			else if (c=='c') false_=(false_==FALSE?TRUE:FALSE); 
			else if (c=='d') debug=TRUE; 
			else if (c=='h') {printHelp(); exit(0);}
			else if (c=='l') {
				int c;
				printf("Errors list\n");
				printf("----------\n");
				for (c=0;c<ERRN;c++) {
					if (c==22) {
						printf("\nWarnings list\n");
						printf("-------------\n");
					}
					printf("#%2d: %s\n",c,ers[c]); 
				}
				exit(0);
			}
			else if (c=='t') isTiming=TRUE;
			else if (c=='v') {printVersion(TRUE); exit(0);}
			else if (c=='x') ef=(ef==TRUE?FALSE:TRUE);
			else if (c=='z') cutOff=TRUE;
			else {
				char t[40];
				sprintf(t,"dib: -%c is an illegal option.\n",c);
				derror_(t);
			}
		}
	}
	
	/* get input file name (discard following names) */ 
	FileName = *argv;

	/* ok; now, if after any option there's no file name left
	 * (or if dib was launched with no parameters), issue an error */
	if (('\0'==*FileName)||(1==argc)) derror_("dib: missing file name.\n");

	/* LOAD program in memory */
	LOAD(FileName);

	/* set up printer position and counter */
	CR();

	/* reset GOSUB limits if in standard mode */
	if (!ef) GOSLIM=1;
	
	/* if extended features are available, ensures that an END statement 
	 * is present in case no END statement is met into LOAD()
	 * Note: this command sets the unreachable line 119881 */
	if (ef && !endn) {
		m[CHUNKS*LIMIT+1]=malloc(4*sizeof(int));
		if (!m[CHUNKS*LIMIT+1]) 
			derror_("dib: out of memory for ultimate END.\n");
		strncpy(m[CHUNKS*LIMIT+1],"END",3);
	}

        /* record starting time if required */
        if (isTiming) {
        	gettimeofday(&tv, NULL);
           	beginS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
        }
    
	/* finally RUN! */
	RUN(1);
    
        /* record ending time if required, and print time report */
        if (isTiming) {
        	gettimeofday(&tv, NULL);
        	endS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
        	endS = (int)(endS - beginS + 0.5);
		/* time report in seconds */
		printf(TIMES, (int)endS);
        }
	
	/* exit: the program gets here only if no error caused 
	 * anticipated exiting in some line other than the last */
	return EXIT_SUCCESS;

} /* end main */

/* end of dib.c */
