/*dib.h
* Copyright (C) 2008-2009 Antonio Maschio @ <tbin at libero dot it>
* 
* This is dib, the Dartmouth 1966 BASIC intepreter, a very naive
* attempt to implement the original Dartmouth College BASIC language as 
* conceived by Kemeny & Kurz in 1964, in its 1966 version (why 1966? 
* I was born in 1966). The interpreter is implemented in a not interactive 
* environment; it is in plain C (for the gcc), and it works as a common 
* UNIX command. See the file "dib_manual.txt" for explanation about the 
* interpreter and the language, and about its history and influences.
*
* This file, in particular, is the header of the main C program. 
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* It's GPL! Enjoy!
*
* (Written with vim, tab stop = 8) */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <sys/time.h>


/* Versioning constants */
#define VERSION "1.0 gamma"
#define BUILD 4
#define AU "Antonio Maschio"		/* Author */
#define EM "tbin at libero dot it"  	/* Email */
#define CY "2008-2009"		  	/* Copyright Year(s) */


/* System constants */
#define LIMIT   999	// dib's memory extension chunk
#define CHUNKS  120	// dib's chunks number
#define STRLIM   74 	// string limit
#define SCRLIM   74	// screen limit
#define FORLIM   26	// FOR limit
#define DATLIM  300	// DATA limit
#define TEMPVAR  24	// temp var for MAT calculations (don't change!)
#define ZONE     15	// width of printing zone
#define MINIZONE  3	// width of printing sub zone
#define TRUE      1
#define FALSE     0
#define MINFF -1E75	// greater negative real
#define INFF   1E75	// greater positive real
#define ZERO  1E-75	// smaller real different from 0
#define FNORMTYPE 0	// normers types: Frobenius norm
#define RNORMTYPE 1	// Row norm
#define CNORMTYPE 2	// Column norm
#define RANDL 32768	// Random generator divisor
#define PSTACK   64	// dimension of the inner RPN stack
#define RADIAN    0	// set radian mode
#define DEGREE    1	// set degree mode


/* INSPECTIVE MACROS - argument is always a char pointer */
// is *p a single variable A..Z
#define issv(p) (isalpha(*(p)) && !isalnum(*((p)+1))) 
// is *p a function
#define isfn(p) (isalpha(*(p)) && isalpha(*((p)+1))) 
// is *p a double variable A0..Z9
#define isdv(p) (isalpha(*(p)) && isdigit(*((p)+1)))
// is *p a number
#define isnm(p) (isdigit(*(p)) || *(p)=='.' || *(p)=='-')
// is *p an array item
#define isav(p) ((issv(p) && *((p)+1)=='(') || (isdv(p) && *((p)+2)=='('))
// is *p a math operator
#define ismo(p) (*(p)=='+'||*(p)=='-'||*(p)=='*'||*(p)=='/'||*(p)=='^'||*(p)=='}'||*(p)=='{')
// is *p a relational operator
#define isro(p) (*(p)=='<'||*(p)=='>'||*(p)=='='||*(p)=='#'||*(p)=='!'||*(p)=='$')
// is *p a logical operator (NOT is a unary operator)
#define islo(p) (*(p)=='&'||*(p)=='|'||*(p)=='_'||*(p)=='%'||*(p)=='~')
// is *p an operator (that is math, relational or logical)
#define isop(p) (ismo(p) || isro(p) || islo(p))
// is *p the comma or the semicolon
#define iscs(p) (*(p)==',' || *(p)==';')


/* DEF structure */
struct t_definition {
	int vars[LIMIT];	// variables as arguments holder
	char* def;		// pointer to the DEF string definition
};
typedef struct t_definition Tdef;


/* DIM structure */
struct t_array {
	int row;
	int col;
	int type; 		// type 0=undeclared, 1=vector, 2=matrix
	double *elem;		// pointer to the array base
};
typedef struct t_array Tarray;


/* error strings 
 * (cfr. 1964 Dartmouth BASIC manual, app. V, pages 53-54) 
 * Note: the Dartmouth compiler acted with one single pass over the entire
 * program before executing it, to check the syntax, tokenize statements and 
 * pack instructions before compiling (cfr. pages 2325 of the 1964 manual).
 * dib, instead, while acting quite identically to the original with
 * respect to the execution of **correct** program lines, differs in the
 * error messages treatment, because it doesn't do any check before execution,
 * but rather **during** execution; some non-blockin error messages are 
 * transparent and let the interpreter go on, printing a message interspersed 
 * with program output. This is a behaviour that won't interfere with well 
 * written programs */
char *ers[]={
	/* Standard errors message as in the 1964 manual for version II */
	"DIMENSION TOO LARGE",	// 0 
	"ILLEGAL CONSTANT",	// 1
	"ILLEGAL FORMULA",	// 2
	"ILLEGAL RELATION",	// 3
	"ILLEGAL LINE NUMBER",  // 4
	"ILLEGAL INSTRUCTION",  // 5
	"ILLEGAL VARIABLE",	// 6
	"INCORRECT FORMAT",	// 7
	"END IS NOT LAST",	// 8
	"NO END INSTRUCTION",   // 9
	"NO DATA",		// 10
	"UNDEFINED FUNCTION",   // 11
	"UNDEFINED NUMBER",	// 12
	"PROGRAM TOO LONG",	// 13 
	"TOO MUCH DATA",	// 14 
	"TOO MANY LABELS",	// 15 
	"TOO MANY LOOPS",	// 16 
	"NOT MATCH WITH FOR",   // 17
	"FOR WITHOUT NEXT",	// 18
	"CUT PROGRAMS OR DIMS", // 19 
	"SUBSCRIPT ERROR",	// 20
	"ILLEGAL RETURN",	// 21
	/* extra non-blocking warnings taken from the Dartmouth BASIC sources
	 * or introduced by myself in a 1964 style */
	"OUT OF DATA",				// 22
	"SQUARE ROOT OF NEGATIVE NUMBER", 	// 23
	"LOG OF NEGATIVE NUMBER",	  	// 24
	"LOG OF ZERO",			  	// 25
	"ABSOLUTE VALUE RAISED TO POWER", 	// 26
	"DIVISION BY ZERO",			// 27
	"",					// 28
	"",					// 29	
	"OVERFLOW",				// 30
	"UNDERFLOW",				// 31
	"ZERO TO A NEGATIVE POWER",		// 32
	"ZERO TO THE ZERO POWER",		// 33
	"GOSUBS NESTED TOO DEEPLY",		// 34
	"ARRAY INDEX MUST BE GREATER THAN ZERO",// 35
	"NO MATRIX INVERSION DONE YET",		// 36
	"IGNORING ONE-ELEMENT LIST OR TABLE",	// 37
	"ASIN OR ACOS ARGUMENT MUST BE BETWEEN 1 AND -1",	// 38
	"ACSH ARGUMENT MUST BE GREATER THAN 1",			// 39
	"COTANGENT OF ZERO",					// 40
	"INVERSION OF A SINGULAR MATRIX IS NOT POSSIBLE",	// 41
	""
};

/* ERRN must be one unity greater than last non-void ERROR */
#define ERRN 42

/* Functions prototypes */
double getVal(void);
double getFunc(void);
int priority(int op);	/* operator priority calculator */
double S(void);		/* the parser */
void G(void);		/* the storer */
void stripBlanks(char *src, char *dest);
int in(char c, char d);
int printNUM(double num);		/* number formatter */
int emitNUM(char minus, char *num);	/* number output */
void doExp(char*n,double*b,int*e);      /* retrieve base and exponent */
void AC(void);		/* Advance Carriage */
void CR(void);		/* Carriage Return  */
int check(char *sb, char *sc);
void perror_(int e, int l, int g);
void perrorEx_(int e, int l);
void derror_(char* t);
void RUN(int sl);
void LOAD(char *f);
void PRINT(void);
void advance(void);		/* currently unused */
void advanceToMinizone(void);	/* currently unused */
void advanceToZone(void); 	/* currently unused */
void fill(int Width);		/* currently unused */
void CHANGE(void);	/* for CHANGE "STRING" TO C, where C is a vector */
void DATA(void);
void DEF(void);		/* DEF with one argument */
void DEFMUL(void);	/* DEF with multiple arguments */
void READ(void);
void RESTORE(void);
void FOR(void);
void NEXT(void);
void GOSUB(void);
void GOTO(void);
void RETURN(void);
void IFTHEN(void);
void ONGOTO(void);
void INPUT(void);
int LET(void);		/* LET : here wrapper for assign/dimAssign*/
int assign(void);	/* assign values to variable */
void DIM(void);
void RANDOMIZE(void);
double getRandom(void);
double getArrayVal(int var, int type, int M, int N);
void setArrayVal(int var, int type, int M, int N, double val);
void dimArray(int var, int type, int M, int N);
void dimAssign(void);	/* assign values to array variables */
int isod(char *p); 	/* is *p a one-dimensional array */
int isaa(char *p);	/* is *p an array assignment like A(X[,Y])=EXPR */
void matReadAssign(int var, int t, int m, int n); /* assign MAT READ values */
void matCopy(int dest, int source); /* MAT C = A */
void MATtype(void);	/* the MAT controller */
void MATREAD(void);	/* MAT READ A(M[,N]),B(M[,N])... */
void MATINPUT(void);	/* MAT INPUT A,B... */
void MATZER(int var, double op); /* MAT C=ZER(M[,N])|CON(M[,N])*/
void MATIDN(int var);   /* MAT C=IDN(M) */
void MATPRINT(void);	/* MAT PRINT A[;|,] */
void MATTRN(int var);	/* MAT C = TRN(A) */
void MATINV(int var);	/* MAT C = INV(A) - wrapper for matinvert() */
void matinvert(int varr, int var1); // inversion and determinant calculation
void MATDET(void);	/* C = DET(A) explicit determinant of square mat. A  */
void MATbyK(int var);	/* MAT C = (exp) * A */
void MATsum(int var, int var1, int var2, int op);	/* MAT C = A +|- B   */
void MATmul(int var, int var1, int var2);		/* MAT C = A * B     */
double MATdot(void);	/* C = DOT(A,B) dot product */
double normer(int var, int type); /* calculator for FNORM, RNORM and CNORM   */
double MATNORM(void);	/* C = [FNORM|RNORM|CNORM](A) - wrapper for normer() */	
double power(double base, double expon);
double GAMMA(void);
double FIX(void);
double dg(double rad);	/* convert from radians to degrees */
double rd(double deg);	/* convert from degrees to radians*/
void printHelp(void);
void printVersion(int flag);
void printWarraty(int flag);
void printConditions(int flag);

/* LOTSA GLOBAL VARIABLES! */

/* l holds the current program line number
 * P is an integer array holding variables contents */
int l;
double P[LIMIT];

/* B is the general input buffer 
 * J is the input buffer used to store DEF formulas */
char B[STRLIM+1], J[STRLIM+1];

/* F is a character-holder used by the blank stripper */
char F[2];

/* m is an array holding pointers to program lines 
 * p is the main char pointer to the current buffer position
 * s and d are various char pointers used into RUN */
//char *m[CHUNKS*LIMIT+10], *p,*q,*x,*y,*z,*s,*d;
char *m[CHUNKS*LIMIT+10], *p,*s,*d;

/* printn is a flag for disabling printing CR after last ';' or ',' 
 * in PRINT statements (1=print CR; 0=don't print CR); printn means print '\n' 
 * counter is the current next printing column on the screen line 
 * isTab is the flag active when a TAB() function is met (used to avoid
 * interpreting the semicolon or the comma after the TAB)
 * cw is a temporary backup value for counter 
 * if cutOff is TRUE (enabled with -z option) all numbers lower than EPSILON 
 * are printed as zero, to avoid cluttered output; 
 * EPSILON contains the customizable limit */
int printn=1;
int counter=0;
int isTab=FALSE;
int cw;
int cutOff=FALSE;
double EPSILON=1E-10;

/* D is an array holding DATA values
 * dc (data counter) yields current pointer to data
 * dataLimit is the total of DATA values found into the listing */
double D[LIMIT];
int dc=0;
int dataLimit=0;

/* fn holds the DEF FN definitions */ 
Tdef fn[LIMIT];

/* Extended features flag 
 * FALSE = default behavior; TRUE = extended features enabled */
int ef=FALSE;

/* END features 
 * endl holds the last line number where a END appears 
 * endn holds the total number of END statements found
 * used into LOAD() */
int endl=0;
int endn=0;


/* GOSUB data management and storing 
 * Note: GOSLIM (=GOSub LIMit) is reset to 1 in main() in default mode
 * isGOSUB is a flag to check if any jump out of a FOR..NEXT cycle is a GOSUB 
 * (and this won't interrupt the cycle) or a GOTO/IF..THEN (and this will 
 * break the cycle) */
int gosubStack[LIMIT];
int gosubTOS = 0;
int GOSLIM=80;
int isGOSUB=FALSE;

/* FOR..NEXT data storing 
 * forcount is a counter for contemporary nested for cycles */
double forStack[6*FORLIM+6];
int forTOS = 0;
int forcount=0;

/* error circuitation & error treatment
 * The error system is built in such a way it stops only with blocking errors; 
 * in other cases, it can safely go on, eventually printing a warning.
 * The flags involved are true_ and false_, which may or not point to TRUE
 * or FALSE; if they are coherent, the system will stop only on meeting
 * blocking errors; if they both point to TRUE, the system will *always*
 * block (this is the default case); to enable partial blocking, 
 * use option -c when starting dib */
int true_= TRUE;
int false_=TRUE;

/* DIM arrays holder */
Tarray dim[LIMIT];

/* debug flag */
int debug=FALSE;

/* ob = OPTION BASE index (0 = index from 0; 1 = index from 1)
 * Default for BASIC I   (1964) ARRAYS = 0; no MAT statements
 * Default for BASIC II  (1964) ARRAYS = 0; MAT ARRAYS = 1 (cardbasic)
 * Default for BASIC III (1966) ARRAYS = 0; MAT ARRAYS = 0
 * Default for BASIC IV  (1967) ARRAYS = 0; MAT ARRAYS = 1
 * Default for BASIC V   (1969) ARRAYS = 0; MAT ARRAYS = 1
 * Default for BASIC VI  (1971) ARRAYS = 0; MAT ARRAYS = 1
 * dib has a consistent base for array indexes, so that weird differentiations
 * of versions II and IV-VI are banned, in default and in extended mode.*/
int ob=0;

/* MAT variables*/
double detValue=0.0; 	// placeholder for DET (MAT INV determinant)
int isdetCalc=FALSE;	// has determinant been calculated by MAT INV routine?
int numValue=0;		// placeholder for NUM (MAT INPUT entries)

/* parser variables */
int opturn=0;		// (don't change!) true when an operator is expected

/* the degree enviroment variable, bu default set on RADIAN */
int degree=RADIAN;

/* Timer string (see October 1964 manual) */
#define TIMES           "\nTIME:   %d SECS.\n"

/* end of dib.h */

